(defvar fep-stack nil)

(defun fep-push ()
  (let ((flag (fep-get-mode nil)))
    (setq fep-stack (cons flag fep-stack))
    (and flag
	 (fep-force-off nil))))

(defun fep-pop ()
  (and (car-safe fep-stack)
       (fep-force-on nil))
  (setq fep-stack (cdr-safe fep-stack)))

(or (fboundp 'si:fep-init)
    (fset 'si:fep-init (symbol-function 'fep-init)))

(defun fep-init ()
  (let (v)
    (setq v (si:fep-init))
    (define-key global-map [kanji] 'toggle-fep-mode)
    (define-key global-map [M-kanji] 'toggle-fep-mode)
    (define-key global-map [C-kanji] 'toggle-fep-mode)
    v))

(defun toggle-fep-mode ()
  (interactive)
  (setq fep-mode (not fep-mode))
  (if fep-mode
      (fep-force-on nil)
    (fep-force-off nil)))

(defun fep-sync ()
  (interactive)
  (setq fep-stack nil)
  (if fep-mode
      (fep-force-on nil)
    (fep-force-off nil)))

(or (fboundp 'si:universal-argument)
    (fset 'si:universal-argument (symbol-function 'universal-argument)))

(defun universal-argument ()
  ""
  (interactive nil)
  (fep-push)
  (unwind-protect
      (si:universal-argument)
    (fep-pop)))

(or (fboundp 'si:quoted-insert)
    (fset 'si:quoted-insert (symbol-function 'quoted-insert)))

(defun quoted-insert (arg)
  ""
  (interactive "*p")
  (fep-push)
  (unwind-protect
      (si:quoted-insert arg)
    (fep-pop)))

(defvar isearch-mode-fep-stack nil)

(defun isearch-mode-fep-push ()
  (let ((flag (fep-get-mode nil)))
    (setq isearch-mode-fep-stack (cons flag isearch-mode-fep-stack))
    (if flag
	(progn
	  (fep-force-off nil)
	  (setq fep-mode nil)))))

(defun isearch-mode-fep-pop ()
  (if (car-safe isearch-mode-fep-stack)
      (progn
	(fep-force-on nil)
	(setq fep-mode t)))
  (setq isearch-mode-fep-stack (cdr-safe isearch-mode-fep-stack)))

(add-hook 'isearch-mode-hook
	  (function
	   (lambda ()
	     (isearch-mode-fep-push))))

(add-hook 'isearch-mode-end-hook
	  (function
	   (lambda ()
	     (isearch-mode-fep-pop))))

(defvar isearch-ime-read-string nil)

(add-hook 'minibuffer-setup-hook
	  (function
	   (lambda ()
	     (and isearch-ime-read-string
		  (fep-force-on nil)))))

(or (fboundp 'si:isearch-fep-read-string)
    (fset 'si:isearch-fep-read-string
	  (symbol-function 'isearch-fep-read-string)))

(defun isearch-fep-read-string ()
  (setq isearch-ime-read-string t)
  (unwind-protect
      (si:isearch-fep-read-string)
    (setq isearch-ime-read-string nil)
    (fep-force-off nil)))


(or (fboundp 'si:y-or-n-p)
    (fset 'si:y-or-n-p (symbol-function 'y-or-n-p)))

(defun y-or-n-p (prompt)
  (fep-push)
  (unwind-protect
      (si:y-or-n-p prompt)
    (fep-pop)))


(or (fboundp 'si:map-y-or-n-p)
    (fset 'si:map-y-or-n-p (symbol-function 'map-y-or-n-p)))

(defun map-y-or-n-p (prompter actor list &optional help action-alist
			      no-cursor-in-echo-area)
  (fep-push)
  (unwind-protect
      (si:map-y-or-n-p prompter actor list help action-alist
		       no-cursor-in-echo-area)
    (fep-pop)))
