/* MS-DOS specific Lisp utilities.  Coded by Manabu Higashida, 1991.
   Major changes May-July 1993 Morten Welinder (only 10% original code left)
   Copyright (C) 1991, 1993 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */


#include <config.h>

#ifdef MSDOS
/* The entire file is within this conditional */

#include <stdio.h>
#include <dos.h>
#ifndef GO32WIN
#include <crt0.h>
#endif
#include "lisp.h"
#include "buffer.h"
#include "termchar.h"
#include "termhooks.h"
#include "frame.h"
#include "commands.h"
#include "dosfns.h"		/* 96.2.3  by K.Fujii */
#include "msdos.h"
#include <go32.h>
#ifdef GO32WIN					/* Go32win Patch '95/02/26 by Y.Matsushima */
#include "go32win.h"
#endif

#ifndef GO32WIN
extern void ScreenReSize (int rows, int cols, int mouse_off);	/* K.Fujii */

DEFUN ("mode25", Fmode25, Smode25, 0, 0, "", "\
Changes the number of rows to 25.")
  ()
{
  union REGS regs;

#ifdef HAVE_X_WINDOWS
  if (!inhibit_window_system)
    return Qnil;
#endif
  if (! ModePCAT ()) return Qnil;

  mouse_off ();
  regs.x.ax = 3;
  int86 (0x10, &regs, &regs);
  regs.x.ax = 0x1101;
  regs.h.bl = 0;
  int86 (0x10, &regs, &regs);
  regs.x.ax = 0x1200;
  regs.h.bl = 32;
  int86 (0x10, &regs, &regs);
  regs.x.ax = 3;
  int86 (0x10, &regs, &regs);
#if 1  /* 95.2.13  by K.Fujii */
  ScreenReSize (ScreenRows (), ScreenCols (), 0);
#else
  Fset_frame_size (Fselected_frame (), ScreenCols (), ScreenRows ());
  Frecenter (Qnil);
  Fredraw_display ();
  if (have_mouse) mouse_init ();
#endif
  return Qnil;
}

DEFUN ("mode4350", Fmode4350, Smode4350, 0, 0, "", "\
Changes the number of rows to 43 (EGA) or 50 (VGA).")
  ()
{
  union REGS regs;

#ifdef HAVE_X_WINDOWS
  if (!inhibit_window_system)
    return Qnil;
#endif
  if (! ModePCAT ()) return Qnil;

  mouse_off ();
  regs.x.ax = 3;
  int86 (0x10, &regs, &regs);
  regs.x.ax = 0x1112;
  regs.h.bl = 0;
  int86 (0x10, &regs, &regs);
  regs.x.ax = 0x1200;
  regs.h.bl = 32;
  int86 (0x10, &regs, &regs);
  regs.x.ax = 0x0100;
  regs.x.cx = 7;
  int86 (0x10, &regs, &regs);
#if 1  /* 95.2.13  by K.Fujii */
  ScreenReSize (ScreenRows (), ScreenCols (), 0);
#else
  Fset_frame_size (Fselected_frame (), ScreenCols (), ScreenRows ());
  Frecenter (Qnil);
  Fredraw_display ();
  if (have_mouse) mouse_init ();
#endif
  return Qnil;
}
#endif /* not GO32WIN */

DEFUN ("int86", Fint86, Sint86, 2, 2, 0,
  "Call specific MSDOS interrupt number INTERRUPT with REGISTERS.\n\
Return the updated REGISTER vector.\n\
\n\
INTERRUPT should be an integer in the range 0 to 255.\n\
REGISTERS should be a vector produced by `make-register' and\n\
`set-register-value'.")
  (intno, regs)
  Lisp_Object intno, regs;
{
  register int i;
  int no;
  union REGS inregs, outregs;
  Lisp_Object val;

  CHECK_NUMBER (intno, 0);
  no = (unsigned long) XINT (intno);
  CHECK_VECTOR (regs, 1);
  if (no < 0 || no > 0xff || XVECTOR (regs)-> size != 8) 
    return Qnil;
  for (i = 0; i < 8; i++)
    CHECK_NUMBER (XVECTOR (regs)->contents[i], 1);

  inregs.x.ax    = (unsigned long) XFASTINT (XVECTOR (regs)->contents[0]);
  inregs.x.bx    = (unsigned long) XFASTINT (XVECTOR (regs)->contents[1]);
  inregs.x.cx    = (unsigned long) XFASTINT (XVECTOR (regs)->contents[2]);
  inregs.x.dx    = (unsigned long) XFASTINT (XVECTOR (regs)->contents[3]);
  inregs.x.si    = (unsigned long) XFASTINT (XVECTOR (regs)->contents[4]);
  inregs.x.di    = (unsigned long) XFASTINT (XVECTOR (regs)->contents[5]);
  inregs.x.cflag = (unsigned long) XFASTINT (XVECTOR (regs)->contents[6]);
  inregs.x.flags = (unsigned long) XFASTINT (XVECTOR (regs)->contents[7]);

  int86 (no, &inregs, &outregs);

  XVECTOR (regs)->contents[0] = make_number (outregs.x.ax);
  XVECTOR (regs)->contents[1] = make_number (outregs.x.bx);
  XVECTOR (regs)->contents[2] = make_number (outregs.x.cx);
  XVECTOR (regs)->contents[3] = make_number (outregs.x.dx);
  XVECTOR (regs)->contents[4] = make_number (outregs.x.si);
  XVECTOR (regs)->contents[5] = make_number (outregs.x.di);
  XVECTOR (regs)->contents[6] = make_number (outregs.x.cflag);
  XVECTOR (regs)->contents[7] = make_number (outregs.x.flags);

  return regs;
}

#ifndef GO32WIN
DEFUN ("msdos-memget", Fdos_memget, Sdos_memget, 2, 2, 0,
  "Read DOS memory at offset ADDRESS into VECTOR.\n\
Return the updated VECTOR.")
  (addr, v)
  Lisp_Object addr, v;
{
  register int i;
  int offs, len;
  char *buf;
  Lisp_Object val;

  CHECK_NUMBER (addr, 0);
  offs = (unsigned long) XINT (addr);
  CHECK_VECTOR (v, 1);
  len = XVECTOR (v)-> size;
  if (len < 1 || len > 2048 || addr < 0 || addr > 0xfffff - len) 
    return Qnil;
  buf = alloca (len);
  dosmemget (offs, len, buf);
  
  for (i = 0; i < len; i++)
    XVECTOR (v)->contents[i] = make_number (buf[i]);

  return v;
}

DEFUN ("msdos-memput", Fdos_memput, Sdos_memput, 2, 2, 0,
  "Write DOS memory at offset ADDRESS from VECTOR.")
  (addr, v)
  Lisp_Object addr, v;
{
  register int i;
  int offs, len;
  char *buf;
  Lisp_Object val;

  CHECK_NUMBER (addr, 0);
  offs = (unsigned long) XINT (addr);
  CHECK_VECTOR (v, 1);
  len = XVECTOR (v)-> size;
  if (len < 1 || len > 2048 || addr < 0 || addr > 0xfffff - len) 
    return Qnil;
  buf = alloca (len);

  for (i = 0; i < len; i++)
    {
      CHECK_NUMBER (XVECTOR (v)->contents[i], 1);
      buf[i] = (unsigned char) XFASTINT (XVECTOR (v)->contents[i]) & 0xFF;
    }

  dosmemput (buf, len, offs);
  return Qt;
}

DEFUN ("msdos-set-keyboard", Fmsdos_set_keyboard, Smsdos_set_keyboard, 1, 2, 0,
  "Set keyboard layout according to COUNTRY.\n\
If the optional argument ALLKEYS is non-nil, the keyboard is mapped for\n\
all keys; otherwise it is only used when the ALT key is pressed.\n\
The current keyboard layout is available in dos-keyboard-code.")
  (country_code, allkeys)
  Lisp_Object country_code;
{
  CHECK_NUMBER (country_code, 0);
  if (!dos_set_keyboard (XINT (country_code), !NILP (allkeys)))
    return Qnil;
  return Qt;
}
#endif /* ! GO32WIN */

#ifndef HAVE_X_WINDOWS
/* Later we might want to control the mouse interface with this function,
   e.g., with respect to non-80 column screen modes.  */

DEFUN ("msdos-mouse-p", Fmsdos_mouse_p, Smsdos_mouse_p, 0, 0, 0, "\
Report whether a mouse is present.")
     ()
{
  if (have_mouse)
    return Qt;
  else
    return Qnil;
}

#ifndef GO32WIN
DEFUN ("set-mouse-position", Fset_mouse_position, Sset_mouse_position, 3, 3, 0,
  "Move the mouse pointer to the center of character cell (X,Y) in FRAME.\n\
WARNING:  If you use this under X windows,\n\
you should call `unfocus-frame' afterwards.")
  (frame, x, y)
     Lisp_Object frame, x, y;
{
  mouse_moveto (XINT (x), XINT (y));
  return Qnil;
}

/* Function to translate colour names to integers.  See lisp/term/pc-win.el
   for its definition.  */


#if 1 /* 96.1.3  by K.Fujii */
struct color_translate {
  char *color_name;
  int color;
};

static struct color_translate msdos_color_translate_tabel[] = {
  /* black 0 */
  { "black",		 0},
  /* blue 1 */
  { "blue",		 1},  { "cadetblue",		 1},
  /* green 2 */
  { "green",		 2},  { "forestgreen",		 2},
  { "darkolivegreen",	 2},
  /* cyan 3 */
  { "cyan",		 3},
  /* red 4 */
  { "red",		 4},  { "firebrick",		 4},
  /* magenta 5 */
  { "magenta",		 5},  { "purple",		 5},
  { "plum1",		 5},
  /* brown 6 */
  { "brown",		 6},  { "darkgoldenrod",	 6},
  { "rosybrown",	 6},
  /* light gray 7 */
  { "lightgray",	 7},  { "light gray",		 7},
  { "lightgrey",	 7},  { "light grey",		 7},
  /* dark gray 8 */
  { "darkgray",		 8},  { "dark gray",		 8},
  { "darkgrey",		 8},  { "dark grey",		 8},
  { "grey40",		 8},  { "grey50",		 8},
  { "gray40",		 8},  { "gray50",		 8},
  { "dimgray",		 8},  { "dimgrey",		 8},
  /* light blue 9 */
  { "lightblue",	 9},  { "light blue",		 9},
  { "royalblue",	 9},  { "skyblue",		 9},
  /* light green 10 */
  { "lightgreen",	10},  { "light green",		10},
  { "olivedrab2",	10},
  /* light cyan 11 */
  { "lightcyan",	11},  { "light cyan",		11},
  /* light red 12 */
  { "lightred",		12},  { "light red",		12},
  { "pink",		12},
  /* light magenta 13 */
  { "lightmagenta",	13},  { "light magenta",	13},
  { "lightpurple",	13},  { "light purple",		13},
  /* yellow 14 */
  { "yellow",		14},  { "goldenrod",		14},
  { "tan",		14},  { "moccasin",		14},
  { "orange",		14},
  /* white 15 */
  { "white",		15},
  { 0,			-1},
};

DEFUN ("msdos-color-translate", Fmsdos_color_translate,
       Smsdos_color_translate, 1, 1, 0, "")
  (name)
     Lisp_Object name;
{
  unsigned char *np;
  Lisp_Object val = Qnil;
  struct color_translate *translate;

  if (NILP (name) || INTEGERP (name)) return name;

  if (SYMBOLP (name))
    {
      if (! NILP (Fboundp (name)))
	{
	  val = Fsymbol_value (name);
	  if (INTEGERP (val))
	    return val;
	  val = Qnil;
	}
      name = Fsymbol_name (name);
    }

  if (STRINGP (name)) {
    np = XSTRING (name)->data;
    for (translate = msdos_color_translate_tabel;
	 translate->color_name; translate++)
      {
	if (! stricmp (np, translate->color_name))
	  {
	    XSETFASTINT (val, translate->color);
	    break;
	  }
      }
  }

  return val;
}
#else
Lisp_Object Qmsdos_color_translate;
#endif

DEFUN ("msdos-mouse-init", Fmsdos_mouse_init, Smsdos_mouse_init, 0, 0, "",
  "Initialize and enable mouse if available.")
  ()
{
  if (have_mouse)
    {
      have_mouse = 1;
      mouse_init ();
      return Qt;
    }
  return Qnil;
}

DEFUN ("msdos-mouse-enable", Fmsdos_mouse_enable, Smsdos_mouse_enable, 0, 0, "",
  "Enable mouse if available.")
  ()
{
  if (have_mouse)
    {
      have_mouse = 1;
      mouse_on ();
    }
  return have_mouse ? Qt : Qnil;
}

DEFUN ("msdos-mouse-disable", Fmsdos_mouse_disable, Smsdos_mouse_disable, 0, 0, "",
  "Disable mouse if available.")
  ()
{
  if (have_mouse) have_mouse = -1;
  return Qnil;
}

DEFUN ("insert-startup-screen", Finsert_startup_screen, Sinsert_startup_screen, 0, 0, "", "\
Insert copy of screen contents prior to starting emacs.\n\
Return nil if startup screen is not available.")
  ()
{
  char *s;
  int rows, cols;
  int i, j;

  if (!dos_get_saved_screen (&s, &rows, &cols))
    return Qnil;
  
  for (i = 0; i < rows; i++)
    {
      for (j = 0; j < cols; j++)
	{
	  insert_char (*s, 1);
	  s += 2;
	}
      insert_char ('\n', 1);
    }

  return Qt;
}
#endif /* ! GO32WIN */
#endif /* ! HAVE_X_WINDOWS */

#ifdef FEPCTRL	/* Demacs 1.1.8 91/11/29 Manabu Higashida */

int InhibitFEPCtrl;

#ifndef GO32WIN  /* 96.2.26  by K.Fujii */
/* patch 95.2.15  by K.Fujii */
int FEPReSizeScreen;
int InhibitFEPReSizeScreen;
/* end of patch */
#endif

DEFUN ("fep-init", Ffep_init, Sfep_init, 0, 0, 0, "")
  ()
{
  extern int fep_init ();

  if (!InhibitFEPCtrl)
    return make_number (fep_init ());
  else
    return Qnil;
}

DEFUN ("fep-term", Ffep_term, Sfep_term, 0, 0, 0, "")
  ()
{
  extern void fep_term ();

  if (!InhibitFEPCtrl)
    fep_term ();
  return Qnil;
}

DEFUN ("fep-on", Ffep_on, Sfep_on, 0, 0, 0, "")
 ()
{
  extern void fep_on ();

  if (!InhibitFEPCtrl
#ifndef GO32WIN  /* 96.2.29  by K.Fujii */
      && ! ModePCAT ()
#endif
      )
    /* patch 95.2.15, 95.10.22  by K.Fujii */
    {
#ifdef GO32WIN
      fep_on ();
#else
      int rows_after, rows_before = ScreenRows ();

      fep_on ();

      if (! InhibitFEPReSizeScreen &&
	  ! sw.fep_status_line)
	{
	  rows_after = ScreenRows ();

	  if (rows_after != rows_before)
	    ScreenReSize (rows_after, 0, 1);
	  else if (FEPReSizeScreen)
	    ScreenReSize (rows_after - 1, 0, 1);
	}
#endif
    }
  /* end of patch */
  return Qnil;
}

DEFUN ("fep-off", Ffep_off, Sfep_off, 0, 0, 0, "")
  ()
{
  extern void fep_off ();

  if (!InhibitFEPCtrl
#ifndef GO32WIN /* 96.2.26  by K.Fujii */
      && ! ModePCAT ()
#endif
      )
    /* patch 95.2.15, 95.10.20  by K.Fujii */
    {
#ifdef GO32WIN /* 96.2.26  by K.Fujii */
      fep_off ();
#else
      int rows_after, rows_before = ScreenRows ();

      fep_off ();

      if (! InhibitFEPReSizeScreen &&
	  ! sw.fep_status_line)
	{
	  rows_after = ScreenRows ();

	  if (rows_after != rows_before ||
	      FEPReSizeScreen)
	    ScreenReSize (rows_after, 0, 1);
	}
#endif
    }
    /* end of patch */
  return Qnil;
}

DEFUN ("fep-force-on", Ffep_force_on, Sfep_force_on, 0, 0, "", "")
  ()
{
  extern void fep_force_on ();

  if (!InhibitFEPCtrl
#ifndef GO32WIN  /* 96.2.26  by K.Fujii */
      && ! ModePCAT ()
#endif
      )
    /* patch 95.2.15, 95.10.20  by K.Fujii */
    {
#ifdef GO32WIN /* 96.2.26  by K.Fujii */
      fep_force_on ();
#else
      int rows_after, rows_before = ScreenRows ();

      fep_force_on ();

      if (! InhibitFEPReSizeScreen &&
	  ! sw.fep_status_line)
	{
	  rows_after = ScreenRows ();

	  if (rows_after != rows_before)
	    ScreenReSize (rows_after, 0, 1);
	  else if (FEPReSizeScreen)
	      ScreenReSize (rows_after - 1, 0, 1);
	}
#endif
    }
    /* end of patch */
  return Qnil;
}

DEFUN ("fep-force-off", Ffep_force_off, Sfep_force_off, 0, 0, "", "")
  ()
{
  extern void fep_force_off ();

  if (!InhibitFEPCtrl
#ifndef GO32WIN  /* 96.2.26  by K.Fujii */
      && ! ModePCAT ()
#endif
      )
    /* patch 95.2.15, 95.10.20  by K.Fujii */
    {
#ifdef GO32WIN /* 96.2.26  by K.Fujii */
      fep_force_off ();
#else
      int rows_after, rows_before = ScreenRows ();

      fep_force_off ();

      if (! InhibitFEPReSizeScreen &&
	  ! sw.fep_status_line)
	{
	  rows_after = ScreenRows ();

	  if (rows_after != rows_before ||
	      FEPReSizeScreen)
	    ScreenReSize (rows_after, 0, 1);
	}
#endif
    }
    /* end of patch */
  return Qnil;
}

DEFUN ("fep-get-mode", Ffep_get_mode, Sfep_get_mode, 0, 0, "", "")
  ()
{
  extern int fep_get_mode ();

  if (!InhibitFEPCtrl
#ifndef GO32WIN /* 96.2.26  by K.Fujii */
      && ! ModePCAT ()
#endif
      )
    return fep_get_mode () ? Qt : Qnil;
  else
    return Qnil;
}

/* patch 96.8.20  by K.Fujii */
DEFUN ("toggle-fep-mode", Ftoggle_fep_mode, Stoggle_fep_mode, 0, 0, "", "")
     ()
{
  if (!InhibitFEPCtrl
#ifndef GO32WIN
      && ! ModePCAT ()
#endif
      ) {
    if (! NILP (current_buffer->fep_mode)) {
      Ffep_force_off ();
      current_buffer->fep_mode = Qnil;
    } else {
      Ffep_force_on ();
      current_buffer->fep_mode = Qt;
    }
  }
  return Qnil;
}
    
#ifndef GO32WIN  /* 96.2.27  by K.Fujii */
DEFUN ("erase-function-key-label", Ferase_function_key_label,
       Serase_function_key_label, 0, 0, "", "")
  ()
{
  if (MachinePc98 () && sw.fep_status_line)
    {
      fwrite ("\033[>1h", 1, 5, stderr);
      fflush (stderr);
      ScreenReSize (ScreenRows (), ScreenCols (), 1);
      sw.fep_status_line = 0;
      InhibitFEPReSizeScreen = 0;
    }
  return Qnil;
}

DEFUN ("display-function-key-label", Fdisplay_function_key_label,
       Sdisplay_function_key_label, 0, 0, "", "")
  ()
{
  if (MachinePc98 () && ! sw.fep_status_line)
    {
      fwrite ("\033[>1l", 1, 5, stderr);
      fflush (stderr);
      ScreenReSize (ScreenRows (), ScreenCols (), 1);
      sw.fep_status_line = 1;
      InhibitFEPReSizeScreen = 1;
    }
  return Qnil;
}

DEFUN ("keep-fep-status-line", Fkeep_fep_status_line,
       Skeep_fep_status_line, 0, 1, "", "")
  (erase_key_label)
     Lisp_Object erase_key_label;
{
  if (ModePCAT ()) return Qnil;

  if (MachinePc98 () && sw.fep_status_line && ! NILP (erase_key_label))
    Ferase_function_key_label ();

  if (! sw.fep_status_line)
    {
      int row;

      row = ScreenRows ();
      if (fep_get_mode () && ! FEPReSizeScreen) row++;
      ScreenReSize (ScreenRows () - 1, ScreenCols (), 1);
      InhibitFEPReSizeScreen = 1;
    }

  return Qnil;
}

DEFUN ("unkeep-fep-status-line", Funkeep_fep_status_line,
       Sunkeep_fep_status_line, 0, 1, "", "")
  (erase_key_label)
     Lisp_Object erase_key_label;
{
  if (ModePCAT ()) return Qnil;

  if (MachinePc98 () && sw.fep_status_line && ! NILP (erase_key_label))
    Ferase_function_key_label ();

  if (! sw.fep_status_line)
    {
      ScreenReSize (ScreenRows (), ScreenCols (), 1);
      InhibitFEPReSizeScreen = 0;
    }

  return Qnil;
}
#endif /* ! GO32WIN */
#endif /* FEPCTRL */

/* patch 94.8.19, 96.2.4  by K.Fujii */
Lisp_Object Qpc98;
Lisp_Object Qibmpc, Qibmpc_us, Qibmpc_jp;
#ifdef GO32WIN					/* Go32win Patch '95/02/26 by Y.Matsushima */
Lisp_Object Qgo32win;
Lisp_Object Qfm, Qtowns;	/* 95.4.4 K.Fujii */
#endif

DEFUN ("dos-machine-type", Fdos_machine_type, Sdos_machine_type, 0, 1,
       "", "")
  (print_flag)
     Lisp_Object print_flag;
{
  Lisp_Object tmp;
  char *p;

#ifdef GO32WIN		/* Go32win Patch '95/02/26 by Y.Matsushima */
			/* 95.4.4 K.Fujii */
  {
    int m = __crt0_mtype >> 4;

    switch (m)
      {
      case 1: tmp = Qpc98; p = "pc98"; break;
      case 2: tmp = Qfm; p = "fm"; break;
      case 3: tmp = Qtowns; p = "towns"; break;
      default: tmp = Qibmpc; p = "ibmpc"; break;
      }
  }
#else
  if (MachinePc98 ())
    tmp = Qpc98, p = "pc98";
  else
    tmp = Qibmpc, p = "ibmpc";
#endif

  if (NILP (print_flag) && INTERACTIVE)
    message ("dos-machine-type: Running machine is %s.", p);

  return tmp;
}

DEFUN ("dos-machine-mode", Fdos_machine_mode, Sdos_machine_mode, 0, 1, "", "")
  (print_flag)
     Lisp_Object print_flag;
{
  Lisp_Object tmp;
  char *p;

#ifdef GO32WIN  /* 95.3.22  by K.Fujii */
  tmp = Qgo32win, p = "go32win";
#else
  switch (__crt0_mtype)
    {
    case __PC98:	tmp = Qpc98;	    p = "pc98";	       break;
    case __DOSV:	tmp = Qibmpc_jp;    p = "ibmpc-jp";    break;
    default:		tmp = Qibmpc_us;    p = "ibmpc-us";    break;
    }
#endif
  if (NILP (print_flag) && INTERACTIVE)
    message ("dos-machine-mode: Running mode is %s.", p);

  return tmp;
}
/* end of patch */

#ifndef GO32WIN  /* 96.2.27  by K.Fujii */
/* patch 94.9.26  by K.Fujii */
extern int load_pc98_font (const char *, int, unsigned short);

Lisp_Object Qpc98_font_bold;
Lisp_Object Qpc98_font_italic;

DEFUN ("si:font-load-pc98", Fsi_font_load_pc98,
       Ssi_font_load_pc98, 3, 3, "", "")
  (font, file, fast)
     Lisp_Object font, file, fast;
{
  Lisp_Object font_path;
  unsigned short font_type;

  if (! MachinePc98 ())
    error ("font-load-pc98: This machine is not `pc98-normal\'.");
  if (strcmp (egetenv ("TERM"), "internal98"))
    error ("font-load-pc98: Now running `internal-terminal\' mode.");

  if (EQ (font, Qpc98_font_bold))
    font_type = 0x7600;
  else if (EQ (font, Qpc98_font_italic))
    font_type = 0x7700;
  else
    error ("font-load-pc98: unknown font.");

  CHECK_STRING (file, 0);

  font_path = Fexpand_file_name (file, Qnil);

  load_pc98_font (XSTRING (font_path)->data,
		  (NILP (fast) ? 0 : 1), font_type);

  return Qnil;
}
#endif /* ! GO32WIN */

/* country info */
int dos_country_code;
int dos_codepage;
int dos_timezone_offset;
int dos_decimal_point;
int dos_keyboard_layout;
unsigned char dos_country_info[DOS_COUNTRY_INFO];

int dos_hyper_key;
int dos_super_key;
int dos_keypad_mode;

Lisp_Object Vdos_version;
Lisp_Object Vdos_display_scancodes;
Lisp_Object Vdos_menubar_clock;
/* patch 96.4.23  by K.Fujii */
Lisp_Object Vdos_menubar_clock_color;
/* end of patch */
Lisp_Object Vdos_timer_hooks;
  
void
init_dosfns ()
{
#ifdef GO32WIN
  union REGS regs;
#else
  _go32_dpmi_seginfo info;
  __dpmi_regs regs;
#endif

#ifndef SYSTEM_MALLOC
  get_lim_data (); /* why the hell isn't this called elsewhere? */
#endif

  regs.x.ax = 0x3000;
#ifdef GO32WIN
  intdos (&regs, &regs);
#else
  __dpmi_int (0x21, &regs);
#endif
  Vdos_version = Fcons (make_number (regs.h.al), make_number (regs.h.ah));

#ifdef GO32WIN					/* Go32win Patch '95/02/26 by Y.Matsushima */
  regs.x.ax = 0x3800;
  regs.x.dx = (unsigned) dos_country_info;
  intdos(&regs, &regs);
  dos_country_code = regs.x.bx;
#else
  /* Obtain the country code by calling Dos via Dpmi.  Don't rely on GO32.  */
  info.size = (sizeof(dos_country_info) + 15) / 16;
  if (_go32_dpmi_allocate_dos_memory (&info))
    dos_country_code = 1;
  else
    {
      regs.x.ax = 0x3800;
      regs.x.ds = info.rm_segment;
      regs.x.dx = 0;
      __dpmi_int (0x21, &regs);
      dos_country_code = regs.x.bx;
      dosmemget (info.rm_segment * 16, DOS_COUNTRY_INFO, dos_country_info);
      _go32_dpmi_free_dos_memory (&info);
    }
  dos_set_keyboard (dos_country_code, 0);
#endif /* ! GO32WIN */

  regs.x.ax = 0x6601;
#ifdef GO32WIN
  intdos (&regs, &regs);
  if (regs.x.cflag)
#else
  __dpmi_int (0x21, &regs);
  if (regs.x.flags & 1)
#endif
    /* Estimate code page from country code */
    switch (dos_country_code) 
      {
      /* patch 96.1.25  by K.Fujii */
      case 81: /* Japan */
	dos_codepage = 932;
	break;
      /* end of patch */
      case 45: /* Denmark */
      case 47: /* Norway */
	dos_codepage = 865;
	break;
      default:
	/* US */
	dos_codepage = 437;
      }
  else
    dos_codepage = regs.x.bx & 0xffff;
}

/*
 *	Define everything
 */
syms_of_dosfns ()
{
#ifndef GO32WIN /* 96.2.27  by K.Fujii */
  defsubr (&Smode25);
  defsubr (&Smode4350);
  defsubr (&Sint86);
  defsubr (&Sdos_memget);
  defsubr (&Sdos_memput);
  defsubr (&Smsdos_mouse_init);
  defsubr (&Smsdos_mouse_enable);
  defsubr (&Smsdos_set_keyboard);
  defsubr (&Sinsert_startup_screen);
  defsubr (&Smsdos_mouse_disable);
#endif
  /* patch 94.8.19, 94.9.26  by K.Fujii */
  defsubr (&Sdos_machine_type);
  defsubr (&Sdos_machine_mode);
#ifndef GO32WIN
  defsubr (&Ssi_font_load_pc98);
#endif
  /* end of patch */
#ifdef FEPCTRL	/* 91.11.29 by M.Higashida */
  defsubr (&Sfep_init);
  defsubr (&Sfep_term);
  defsubr (&Sfep_on);
  defsubr (&Sfep_off);
  defsubr (&Sfep_force_on);
  defsubr (&Sfep_force_off);
  defsubr (&Sfep_get_mode);
  defsubr (&Stoggle_fep_mode);
#ifndef GO32WIN  /* 96.2.27  by K.Fujii */
  /* patch 95.2.16  by K.Fujii */
  defsubr (&Skeep_fep_status_line);
  defsubr (&Sunkeep_fep_status_line);
  defsubr (&Serase_function_key_label);
  defsubr (&Sdisplay_function_key_label);
#endif
  /* end of patch */
#endif /* FEPCTRL */
#if !defined (HAVE_X_WINDOWS) && !defined (GO32WIN)
  defsubr (&Smsdos_mouse_p);
  defsubr (&Sset_mouse_position);

#if 1 /* 96.1.3  by K.Fujii */
  defsubr (&Smsdos_color_translate);
#else
  Qmsdos_color_translate = intern ("msdos-color-translate");
  staticpro (&Qmsdos_color_translate);
#endif
#endif /* ! HAVE_X_WINDOWS && ! GO32WIN */

  DEFVAR_INT ("dos-country-code", &dos_country_code,
    "The country code returned by Dos when Emacs was started.\n\
Usually this is the international telephone prefix.");

  DEFVAR_INT ("dos-codepage", &dos_codepage,
    "The codepage active when Emacs was started.\n\
The following are known:\n\
	437	United States\n\
	850	Multilingual (Latin I)\n\
	852	Slavic (Latin II)\n\
	857	Turkish\n\
	860	Portugal\n\
	861	Iceland\n\
	863	Canada (French)\n\
	865	Norway/Denmark");

  DEFVAR_INT ("dos-timezone-offset", &dos_timezone_offset,
    "The current timezone offset to UTC in minutes.
Implicitly modified when the TZ variable is changed.");
  
  DEFVAR_LISP ("dos-version", &Vdos_version,
    "The (MAJOR . MINOR) Dos version (subject to modification with setver).");

  DEFVAR_LISP ("dos-display-scancodes", &Vdos_display_scancodes,
    "*When non-nil, the keyboard scan-codes are displayed at the bottom right\n\
corner of the display (typically at the end of the mode line).\n\
The output format is: scan code:char code*modifiers.");
  Vdos_display_scancodes = Qnil;
  
  DEFVAR_LISP ("dos-menubar-clock", &Vdos_menubar_clock,
    "*When non-nil, the current time is displayed in the upper right\n\
corner of the screen (typically at the end of the menu bar).");
  Vdos_menubar_clock = Qt;

  /* patch 96.4.23  by K.Fujii */
  DEFVAR_LISP ("dos-menubar-clock-color", &Vdos_menubar_clock_color,
    "Keep menubar clock displayed foreground and background color.\n
if nil is used 'default' face, else alist (foregroung-color . background-color).");
  Vdos_menubar_clock_color = Qnil;
  /* end of patch */

  DEFVAR_LISP ("dos-timer-hooks", &Vdos_timer_hooks,
    "List of hooks which are run every second.");
  Vdos_timer_hooks = Qnil;
  
  DEFVAR_INT ("dos-hyper-key", &dos_hyper_key,
    "*If set to 1, use right ALT key as hyper key.\n\
If set to 2, use right CTRL key as hyper key.");
  dos_hyper_key = 0;
  
  DEFVAR_INT ("dos-super-key", &dos_super_key,
    "*If set to 1, use right ALT key as super key.\n\
If set to 2, use right CTRL key as super key.");
  dos_super_key = 0;
  
  DEFVAR_INT ("dos-keypad-mode", &dos_keypad_mode,
    "*Controls what key code is returned by a key in the numeric keypad.\n\
The `numlock ON' action is only taken if no modifier keys are pressed.\n\
The value is an integer constructed by adding the following bits together:\n\
 \n\
  0x00	Digit key returns digit    (if numlock ON)\n\
  0x01	Digit key returns kp-digit (if numlock ON)\n\
  0x02	Digit key returns M-digit  (if numlock ON)\n\
  0x03	Digit key returns edit key (if numlock ON)\n\
 \n\
  0x00	Grey key returns char      (if numlock ON)\n\
  0x04	Grey key returns kp-key    (if numlock ON)\n\
 \n\
  0x00	Digit key returns digit    (if numlock OFF)\n\
  0x10	Digit key returns kp-digit (if numlock OFF)\n\
  0x20	Digit key returns M-digit  (if numlock OFF)\n\
  0x30	Digit key returns edit key (if numlock OFF)\n\
 \n\
  0x00	Grey key returns char      (if numlock OFF)\n\
  0x40	Grey key returns kp-key    (if numlock OFF)\n\
 \n\
  0x200	ALT-0..ALT-9 in top-row produces shifted codes.");
  dos_keypad_mode = 0x75;
  
  DEFVAR_INT ("dos-keyboard-layout", &dos_keyboard_layout,
    "Contains the country code for the current keyboard layout.\n\
Use msdos-set-keyboard to select another keyboard layout.");
  dos_keyboard_layout = 1;	/* US */
  
  DEFVAR_INT ("dos-decimal-point", &dos_decimal_point,
    "If non-zero, it contains the character to be returned when the\n\
decimal point key in the numeric keypad is pressed when Num Lock is on.\n\
If zero, the decimal point key returns the country code specific value.");
  dos_decimal_point = 0;
#ifdef GO32WIN					/* Go32win Patch '95/02/26 by Y.Matsushima */
  Qgo32win = intern("go32win");
  staticpro (&Qgo32win);	/* 95.3.22 K.Fujii */
  Qfm = intern ("fm");
  staticpro (&Qfm);
  Qtowns = intern ("towns");
  staticpro (&Qtowns);
#endif
/* patch 94.8.19, 94.9.26  by K.Fujii */
  Qpc98 = intern ("pc98");
  staticpro (&Qpc98);
  Qibmpc = intern ("ibmpc");
  staticpro (&Qibmpc);
  Qibmpc_us = intern ("ibmpc-us");
  staticpro (&Qibmpc_us);
  Qibmpc_jp = intern ("ibmpc-jp");
  staticpro (&Qibmpc_jp);
#ifndef GO32WIN
  Qpc98_font_bold = intern ("pc98-font-bold");
  staticpro (&Qpc98_font_bold);
  Qpc98_font_italic = intern ("pc98-font-italic");
  staticpro (&Qpc98_font_italic);
#endif
/* end of patch */

#ifdef FEPCTRL /* 91.11.29 M.Higashida */
  DEFVAR_BOOL ("inhibit-fep-control", &InhibitFEPCtrl,
    "*Non-nil means that Kana-Kanji FEP control is inhibited.");
  InhibitFEPCtrl = 1;
  /* patch 95.2.16  by K.Fujii */
#ifndef GO32WIN
  DEFVAR_BOOL ("fep-resize-screen", &FEPReSizeScreen, "");
  FEPReSizeScreen = 0;
  DEFVAR_BOOL ("inhibit-fep-resize-screen",
	       &InhibitFEPReSizeScreen,"");
  InhibitFEPReSizeScreen = 1;
  /* end of patch */
#endif
#endif
}
#endif /* MSDOS */

