#ifndef _DOSTERM_H_
#define _DOSTERM_H_

/* Dummy types.  */
typedef int GC;
typedef int Pixmap;
typedef int Display;
typedef int Window;
#define PIX_TYPE int
#define XDISPLAY

#define IT_FONT_DEFAULT 0
#define IT_FONT_BOLD 1
#define IT_FONT_ITALIC 2
#define IT_FONT_BOLD_ITALIC 3
#define IT_FONT_UNDERLINE 8
#define IT_FONT_MAX_ 5
#define IT_FONT_MAX 32

typedef struct __IT_font
{
  unsigned char *internal;
  union
    {
      unsigned char *alias;
      int id;
    } u;
} XFontStruct;

extern XFontStruct IT_font_tab[IT_FONT_MAX];

#ifdef GO32WIN  /* 96.11.26  by K.Fujii */
#define stwin_face_table XFontStruct
#define win_face_table IT_font_tab
#endif

/* The structure representing a frame.

   We declare this even if MULTI_FRAME is not defined, because when
   we lack multi-frame support, we use one instance of this structure
   to represent the one frame we support.  This is cleaner than
   having miscellaneous random variables scattered about.  */

enum output_method {
  output_termcap,
  output_x_window,
  output_msdos_raw,
  output_go32win,
  output_win32
};

#ifdef TERMINAL_FACE
struct f_display		/* dummy display structure to access face */
{
 /* Table of parameter faces for this frame.  Any X resources (pixel
     values, fonts) referred to here have been allocated explicitly
     for this face, and should be freed if we change the face.  */
  struct face **param_faces;
  int n_param_faces;

  /* Table of computed faces for this frame.  These are the faces
     whose indexes go into the upper bits of a glyph, computed by
     combining the parameter faces specified by overlays, text
     properties, and what have you.  The X resources mentioned here
     are all shared with parameter faces.  */
  struct face **computed_faces;
  int n_computed_faces;		/* How many are valid */
  int size_computed_faces;	/* How many are allocated */
};

/* termcap driven display */
struct t_display
{
  struct f_display f;
};
#endif /* TERMINAL_FACE */

/* This is a cut-down version of the one in xterm.h, which see.  */
struct x_output
{
#ifdef TERMINAL_FACE		/* face is common object for terminal and x */
  struct f_display f;
#endif /* TERMINAL_FACE */
  int left_pos;
  int top_pos;
  int line_height;
  PIX_TYPE background_pixel;
  PIX_TYPE foreground_pixel;
  XFontStruct *font;
#ifndef TERMINAL_FACE
  struct face **param_faces;
  int n_param_faces;
  struct face **computed_faces;
  int n_computed_faces;
  int size_computed_faces;
#endif
};

extern struct x_output the_only_x_display;
extern Display *x_current_display;

#ifdef TERMINAL_FACE
#define FRAME_PARAM_FACES(_f) (the_only_x_display.f.param_faces)
#define FRAME_N_PARAM_FACES(_f) (the_only_x_display.f.n_param_faces)
#define FRAME_COMPUTED_FACES(_f) (the_only_x_display.f.computed_faces)
#define FRAME_N_COMPUTED_FACES(_f) (the_only_x_display.f.n_computed_faces)
#define FRAME_SIZE_COMPUTED_FACES(_f) (the_only_x_display.f.size_computed_faces)
#define FRAME_DEFAULT_FACE(_f) (the_only_x_display.f.computed_faces[0])
#define FRAME_MODE_LINE_FACE(_f) (the_only_x_display.f.computed_faces[1])
#else /* not TERMINAL_FACE */
#define FRAME_PARAM_FACES(_f) (the_only_x_display.param_faces)
#define FRAME_N_PARAM_FACES(_f) (the_only_x_display.n_param_faces)
#define FRAME_COMPUTED_FACES(_f) (the_only_x_display.computed_faces)
#define FRAME_N_COMPUTED_FACES(_f) (the_only_x_display.n_computed_faces)
#define FRAME_SIZE_COMPUTED_FACES(_f) (the_only_x_display.size_computed_faces)
#define FRAME_DEFAULT_FACE(_f) (the_only_x_display.computed_faces[0])
#define FRAME_MODE_LINE_FACE(_f) (the_only_x_display.computed_faces[1])
#endif
#define FRAME_DEFAULT_PARAM_FACE(_f) (FRAME_PARAM_FACES (f)[0])
#define FRAME_MODE_LINE_PARAM_FACE(_f) (FRAME_PARAM_FACES (f)[1])
#define FRAME_X_DISPLAY(_f) ((Display *) 0)
#define FRAME_FOREGROUND_PIXEL(_f) (the_only_x_display.foreground_pixel)
#define FRAME_BACKGROUND_PIXEL(_f) (the_only_x_display.background_pixel)
#define FRAME_FONT(_f) (the_only_x_display.font)

#define XFreeGC (void)
#define same_size_fonts(foo,bar) (1)
#define unload_font(p1,p2)
#define unload_color(p1,p2)
#define x_destroy_bitmap(p1,p2)
#define load_pixmap(p1,p2,p3,p4) (0)
#define XGetGeometry(p1,p2,p3,p4,p5,p6,p7,p8,p9)
#define DisplayWidth(p1,p2) (the_only_frame.width)
#define DisplayHeight(p1,p2) (the_only_frame.height)
#define XMenuSetAEQ (void)
#define XMenuSetFreeze (void)
#define XMenuRecompute (void)
#define FONT_WIDTH(foo) 1
#define FONT_HEIGHT(foo) 1

/* Function `getcbrk' is the most harmless I can think of right now...  */
#define check_x if (FRAME_TERMCAP_P (bar)) error ("Not running under a windows system."); else getcbrk
#define x_mouse_leave getcbrk

#define XM_FAILURE -1
#define XM_SUCCESS 1
#define XM_NO_SELECT 2
#define XM_IA_SELECT 3
#define ButtonReleaseMask 0

typedef struct x_menu_struct
{
  int count;
  unsigned char **text;
  struct x_menu_struct **submenu;
  int *panenumber; /* Also used as enable.  */
  int allocated;
  int panecount;
  int width;
} XMenu;

XMenu *XMenuCreate (Display *, Window, char *);
int XMenuAddPane (Display *, XMenu *, char *, int);
int XMenuAddSelection (Display *, XMenu *, int, int, char *, int);
void XMenuLocate (Display *, XMenu *, int, int, int, int,
		  int *, int *, int *, int *);
int XMenuActivate (Display *, XMenu *, int *, int *, int, int, unsigned, char **);
void XMenuDestroy (Display *, XMenu *);

#endif /* _DOSTERM_H_ */
