/* screen.h */

#ifndef __SCREEN_H__
#define __SCREEN_H__
#include <dpmi.h>

typedef struct _screen {
  int nrow;			/* Screen Rows */
  int ncol;			/* Screen Cols */
  int _screen_size;		/* sw.ncol * sw.nrow */
  int old_y, old_x;		/* old corsor position in screen */
  int y, x;			/* cursor position in screen */
  int sy, sx;			/* save cursor position in screen */
  int _y_, _x_;			/* work corsor position */
  unsigned char save_attrib;
  unsigned char attrib;
  int scroll_start, scroll_end;	/* scroll erea start and end */
  int fep_status_line;		/* flag fep-status-line and function
				   key label */
  struct 
    {
      _go32_dpmi_seginfo info;
      unsigned pm_size;
      unsigned output_string_rm_buffer;
    }
  work_buffer;
} Screen;

extern Screen sw;
#define ScreenPrimary _go32_info_block.linear_address_of_primary_screen
#define ScreenSecondary _go32_info_block.linear_address_of_secondary_screen
extern int print_ascii_in_jis;
extern int pc98_bold_flag;
extern int pc98_italic_flag;

int load_pc98_font (const char *p, int fast, unsigned short font_code);

#define ScreenAttrib sw.attrib
#define ScreenFepStatusLine sw.fep_status_line
#define ScreenX sw.x
#define ScreenY sw.y

#define ScreenRmWorkBuffer() (sw.work_buffer.output_string_rm_buffer)
#define ScreenWorkBufferRmSegment() (sw.work_buffer.info.rm_segment)
#define ScreenWorkBufferRmOffset() (0)
#define ScreenGetRows() (sw.nrow)
#define ScreenSetRows(r) sw.nrow = (r)
#define ScreenGetCols() (sw.ncol)
#define ScreenSetCols(c) sw.ncol = (c)
#define ScreenSetPos(_y,_x)				\
  do {							\
    sw._y_ = _y;					\
    sw._x_ = _x;					\
    if (sw._y_ < 0) sw._y_ = 0;				\
    else if (sw._y_ >= sw.nrow) sw._y_ = sw.nrow - 1;	\
    if (sw._x_ < 0) sw._x_ = 0;				\
    else if (sw._x_ >= sw.ncol) sw._x_ = sw.ncol - 1;	\
    if (sw.y != sw._y_ || sw.x != sw._x_) {		\
      sw.old_y = sw.y; sw.old_x = sw.x;			\
      sw.y = sw._y_; sw.x = sw._x_;			\
      ScreenSetCursor (sw.y, sw.x);			\
    }							\
  } while (0)

#define ScreenGetPosX() (sw.x)
#define ScreenSetPosX(_x)				\
  do {							\
    sw._x_ = _x;					\
    if (sw._x_ < 0) sw._x_ = 0;				\
    else if (sw._x_ >= sw.ncol) sw._x_ = sw.ncol - 1;	\
    if (sw.x != sw._x_) {				\
      sw.old_x = sw.x;					\
      sw.x = sw._x_;					\
      ScreenSetCursor (sw.y, sw.x);			\
    }							\
  } while (0)

#define ScreenGetPosY() (sw.y)
#define ScreenSetPosY(_y)				\
  do {							\
    sw._y_ = _y;					\
    if (sw._y_ < 0) sw._y_ = 0;				\
    else if (sw._y_ >= sw.nrow) sw._y_ = sw.nrow - 1;	\
    if (sw.y != sw._y_) {				\
      sw.old_y = sw.y;					\
      sw.y = sw._y_;					\
      ScreenSetCursor (sw.y, sw.x);			\
    }							\
  } while (0)

#define ScreenIncPosX()					\
  do {							\
    sw._x_ = sw.x + 1;					\
    if (sw._x_ < sw.ncol) {				\
      sw.old_x = sw.x;					\
      sw.x = sw._x_;					\
      ScreenSetCursor (sw.y, sw.x);			\
    }							\
  } while (0)

#define ScreenIncPosY()					\
  do {							\
    sw._y_ = sw.y + 1;					\
    if (sw._y_ < sw.nrow) {				\
      sw.old_y = sw.y;					\
      sw.y = sw._y_;					\
      ScreenSetCursor (sw.y, sw.x);			\
    }							\
  } while (0)

#define ScreenDecPosX()					\
  do {							\
    sw._x_ = sw.x - 1;					\
    if (sw._x_ >= 0) {					\
      sw.old_x = sw.x;					\
      sw.x = sw_x_;					\
      ScreenSetCursor (sw.y, sw.x)			\
    }							\
  } while (0)

#define ScreenDecPosY()					\
  do {							\
    sw._y_ = sw.y - 1;					\
    if (sw._y_ >= 0) {					\
      sw.old_y = sw.y;					\
      sw.y = sw._y_;					\
      ScreenSetCursor (sw.y, sw.x);			\
    }							\
  } while (0)

#define ScreenForwardX(_x)				\
  do {							\
    sw._x_ = sw.x + _x;					\
    if (sw._x_ >= sw.ncol) sw._x_ = sw.ncol - 1;	\
    if (sw.x != sw._x_) {				\
      sw.old_x = sw.x;					\
      sw.x = sw._x_;					\
      ScreenSetCursor (sw.y, sw.x);			\
    }							\
  } while (0)

#define ScreenBackwardX(_x)				\
  do {							\
    sw._x_ = sw.x - _x;					\
    if (sw._x_ < 0) sw._x_ = 0;				\
    if (sw.x != sw._x_) {				\
      sw.old_x = sw.x;					\
      sw.x = sw._x_;					\
      ScreenSEtCursor (sw.y, sw.x);			\
    }							\
  } while ()

#define ScreenNextY(_y)					\
  do {							\
    sw._y_ = sw.y + _y;					\
    if (sw._y_ >= sw.nrow) sw._y_ = sw.nrow - 1;	\
    if (sw.y != sw._y_) {				\
      sw.old_y = sw.y;					\
      sw.y = sw._y_;					\
      ScreenSetCursor (sw.y, sw.x);			\
    }							\
  } while (0)

#define ScreenPrevY(_y)					\
  do {							\
    sw._y_ = sw.y - _y;					\
    if (sw._y_ < 0) sw._y_ = 0;				\
    if (sw.y != sw._y_) {				\
      sw.old_y = sw.y;					\
      sw.y = sw._y_;					\
      ScreenSetCursor (sw.y, sw.x);			\
    }							\
  } while (0)

#define ScreenSavePos(a) sw.sy = sw.y, sw.sx = sw.x; if (a) sw.save_attrib = sw.attrib
#define ScreenLoadPos(a)				\
  do {							\
    sw._y_ = sw.sy; sw._x_ = sw.sx;			\
    sw.sy = sw.sx = 0;					\
    if (sw.x != sw._x_ || sw.y != sw._y_) {		\
      sw.old_y = sw.y; sw.old_x = sw.x;			\
      sw.y = sw._y_; sw.x = sw._x_;			\
      ScreenSetCursor (sw.y, sw.x);			\
    }							\
  } while (0)

#define ScreenResetPos() sw.x = sw.sx = sw.y = sw.sy = sw.scroll_start = 0, sw.scroll_end = sw.nrow - 1
#define ScreenScrollStart() sw.scroll_start
#define ScreenScrollEnd() sw.scroll_end
#define ScreenSetScrollStart(s) sw.scroll_start = (s)
#define ScreenSetScrollEnd(e) sw.scroll_end = (e)
#define ScreenGetSize() (sw._screen_size)

#ifdef __GNUC__
static __inline__ void ScreenScrollRange (int top, int botm)
{
  if (top < 0) top = 0;
  if (top >= sw.nrow) top = sw.nrow - 1;
  if (botm < 0) botm = 0;
  if (botm >= sw.nrow) botm = sw.nrow - 1;
  if (top > botm)
    {
      int tmp = botm;
      botm = top;
      top = tmp;
    }
  sw.scroll_start = top;
  sw.scroll_end = botm;
}

#else
#define ScreenScrollRange(t, b)				\
do {							\
    int top, botm, tmp;					\
    top = (t), botm = (b);				\
    if (top < 0) top = 0;				\
    if (botm < 0) botm = 0;				\
    if (top >= sw.nrow) top = sw.nrow - 1;		\
    if (botm >= sw.nrow) botm = sw.nrow - 1;		\
    if (top > botm)					\
      tmp = botm, botm = top, top = tmp;		\
    sw.scroll_start = top, sw.scroll_end = botm;	\
  while (0)
#endif

void ScreenInit (int defattr, int native98);
void ScreenTerminate (void);
void ScreenSetCursor (int row, int col);
void ScreenGetCursor (int *row, int *col);
int ScreenRows (void);
int ScreenCols (void);
int ScreenMode (void);
void ScreenClear (void);
void ScreenPutChar (int ch, int attr, int x, int y);
void ScreenUpdate (void *screenbuffer);
void ScreenUpdateLine (void *screenline, int row);
void ScreenRetrieve (void *screenbuffer);
void _ScreenScrollUp (int top, int botm, int lines);
#define ScreenScrollUp(l) _ScreenScrollUp (sw.scroll_start, sw.scroll_end, (l))
void _ScreenScrollDown (int top, int botm, int lines);
#define ScreenScrollDown(l) _ScreenScrollDown (sw.scroll_start, sw.scroll_end, (l))
void ScreenOutputString (int x, int y, unsigned char *s, int c, unsigned char a);
#endif /* __SCREEN_H__ */
