/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
#include <libc/stubs.h>
#include <stdio.h>		/* For FILENAME_MAX */
#include <stdlib.h>
#include <errno.h>		/* For errno */
#include <ctype.h>		/* For tolower */
#include <string.h>		/* For strlen() */
#include <fcntl.h>
#include <go32.h>
#include <dpmi.h>		/* FOR dpmisim */
#include <crt0.h>
#include <sys/stat.h>
#include <libc/dosio.h>
#ifdef emacs
#include <jctype.h>
#else
#define iskanji(c) (((unsigned char)(c) >= 0x81 && \
		     (unsigned char)(c) <= 0x9f) || \
		    ((unsigned char)(c) >= 0xe0 && \
		     (unsigned char)(c) <= 0xfc))
#define iskanji2(C) ((unsigned char)(c) >= 0x40 && \
		     (unsigned char)(c) <= 0xfc && \
		     (unsigned char)(c) != 0x7f)
#endif

static unsigned use_lfn;

static unsigned char *__get_current_directory(unsigned char *out, int drive_number);

static unsigned char *
__get_current_directory(unsigned char *out, int drive_number)
{
  __dpmi_regs r;
  unsigned char tmpbuf[FILENAME_MAX];

  *out++ = '/';		/* 96.12.12  by K.Fujii */
  *out   = '\0';
  memset(&r, 0, sizeof(r));
  if(use_lfn)
    r.x.ax = 0x7147;
  else
    r.h.ah = 0x47;
  r.h.dl = drive_number + 1;
  r.x.si = __tb_offset;
  r.x.ds = __tb_segment;
  __dpmi_int(0x21, &r);

  if (r.x.flags & 1)
  {
    errno = r.x.ax;
    return out;
  }
  else
  {
    unsigned char *sp, *dp;

    dosmemget(__tb, sizeof(tmpbuf), tmpbuf);
    {
      unsigned c;

      dp = out /* + 1 */;
      sp = tmpbuf;

      while (*sp) {
	c = *sp++;
	if (iskanji (c) && iskanji2 (*sp)) {
	  *dp++ = c;
	  c = *sp++;
	}
	else if (c == '\\')
	  c = '/';
	*dp++ = c;
      }
      *dp = '\0';
    }
#if 0
    /* Root path, don't insert "/", it'll be added later */
    if (*(out + 1) != '\0')
      *out = '/';
    else
      *out = '\0';
    return out + strlen(out);
#else
    return dp;
#endif
  }
}

__inline__ static int
is_slash(int c)
{
  return c == '/' || c == '\\';
}

__inline__ static int
is_term(int c)
{
  return c == '/' || c == '\\' || c == '\0';
}

/* Takes as input an arbitrary path.  Fixes up the path by:
   1. Removing consecutive slashes
   2. Removing trailing slashes
   3. Making the path absolute if it wasn't already
   4. Removing "." in the path
   5. Removing ".." entries in the path (and the directory above them)
   6. Adding a drive specification if one wasn't there
   7. Converting all slashes to '/'
 */
void
_fixpath(const char *in, char *out)
{
  int		drive_number;
  const unsigned char *ip = (unsigned char *)in;
  unsigned char	*op = (unsigned char *)out;
  int		preserve_case = _preserve_fncase();
  unsigned char *name_start;
  /* patch 96.3.20  by K.Fujii */
  int length = 8;
  int suffix = 0;
  /* end of patch */
  use_lfn = _USE_LFN;

  /* Add drive specification to output string */
  if (((*ip >= 'a' && *ip <= 'z') ||
       (*ip >= 'A' && *ip <= 'Z'))
      && (*(ip + 1) == ':'))
  {
    if (*ip >= 'a' && *ip <= 'z') {
      drive_number = *ip - 'a';
      *op++ = *ip++;
    } else {
      drive_number = *ip - 'A';
      if (*ip <= 'Z')
	*op++ = drive_number + 'a';
      else
	*op++ = *ip;
      ip++;
    }
    *op++ = *ip++;
  }
  else
  {
    __dpmi_regs r;
    r.h.ah = 0x19;
    __dpmi_int(0x21, &r);
    drive_number = r.h.al;
    *op++ = drive_number + (drive_number < 26 ? 'a' : 'A');
    *op++ = ':';
  }

  /* Convert relative path to absolute */
  if (!is_slash(*ip))
    op = __get_current_directory(op, drive_number);

  /* Step through the input path */
  while (*ip)
  {
    /* Skip input slashes */
    if (is_slash(*ip))
    {
      ip++;
      /* patch 96.3.20  by K.Fujii */
      *op++ = '/';
      length = 8;
      /* end of patch */
     }
    /* patch 96.3.20  by K.Fujii */
    else if (*ip == '.') {
      ip++;
      /* Skip "." and output nothing */
      if (is_term (*(ip))) {
	continue;
      }
      /* Skip ".." and remove previous output directory */
      else if (*(ip) == '.' && is_term(*(ip + 1)))
	{
	  ip++;
	  /* Don't back up over drive spec */
	  if (op > (unsigned char *)out + 2) {
	    /* This requires "/" to follow drive spec */
	    /* patch 96.3.2  by K.Fujii */
	    while (op > (unsigned char *)out && ! is_slash (*--op))
	      ;
	    op++;
	  }
	  continue;
	}
      length = 3;
      *op++ = '.';
      suffix = 1;
    }
#if 0 /* 96.3.20  by K.Fujii */
    /* Copy path component from in to out */
    *op++ = '/';
#endif
    /* patch 96.3.2  by K.Fujii */
    if (! use_lfn) {
      while (! is_term (*ip)) {
	if (*ip == '.') {
	  if (suffix) {
	    length = 0;
	    goto xxyy;
	  }
	  else break;
	}
	if (iskanji (*ip) && iskanji2 (*(ip + 1))) {
	  if (length > 1) *op++ = *ip;
	  ip++;
	  length--;
	}
	if (length > 0)
	  *op++ = *ip;
	length--;
      xxyy:
	ip++;
      }
      suffix = 0;
    } else {
      while (! is_term (*ip) && *ip != '.') {
	if (iskanji (*ip) && iskanji2 (*(ip + 1)))
	  *op++ = *ip++;
	*op++ = *ip++;
      }
    }
    /* end of patch */
  }

  /* If root directory, insert trailing slash */
  if (op == (unsigned char *)out + 2) *op++ = '/';

  /* Null terminate the output */
  *op = '\0';

  /* switch FOO\BAR to foo/bar, downcase where appropriate */
  for (op = out + 3, name_start = op - 1; *name_start; op++)
  {
    char long_name[FILENAME_MAX], short_name[13];

    if (*op == '\\')
      *op = '/';
    if (!preserve_case && (*op == '/' || *op == '\0'))
    {
      memcpy(long_name, name_start+1, op - name_start - 1);
      long_name[op - name_start - 1] = '\0';
      if (!strcmp((char *)_lfn_gen_short_fname(long_name, short_name), long_name))
      {
	while (++name_start < op)
	  if (*name_start >= 'A' && *name_start <= 'Z')
	    *name_start += 'a' - 'A';
      }
      else
	name_start = op;
    }
    else if (*op == '\0')
      break;
  }
}

