/***************************************
  Win32 Window System Fontset Manager
        Written by himi
***************************************/
#include <config.h>
#include "lisp.h"
#include "w32term.h"
#include "w32fontset.h"
#include "frame.h"
#include "dispextern.h"
#include "coding.h"
#include "charset.h"
#include <commdlg.h>

int fontset_num;
Lisp_Object Vwin32_system_coding_system;
Lisp_Object Vwin32_fontset;
static win32_fontset **win32_fontset_table = 0;

#if 1
static LOGFONT default_normal_font =
{
  0, 0, 0, 0, FW_NORMAL, FALSE, FALSE, FALSE, DEFAULT_CHARSET,
  OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
  FF_DONTCARE | FIXED_PITCH, (unsigned char*)"FixedSys"
};

static LOGFONT default_italic_font =
{
  0, 0, 0, 0, FW_NORMAL, TRUE, FALSE, FALSE, DEFAULT_CHARSET,
  OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
  FF_DONTCARE | FIXED_PITCH, (unsigned char*)"FixedSys"
};

static LOGFONT default_bold_font =
{
  0, 0, 0, 0, FW_BOLD, FALSE, FALSE, FALSE, DEFAULT_CHARSET,
  OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
  FF_DONTCARE | FIXED_PITCH, (unsigned char*)"FixedSys"
};

static LOGFONT default_italic_bold_font =
{
  0, 0, 0, 0, FW_BOLD, TRUE, FALSE, FALSE, DEFAULT_CHARSET,
  OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
  FF_DONTCARE | FIXED_PITCH, (unsigned char*)"FixedSys"
};
#endif

LPLOGFONT
get_logfont(win32_fontset *fs, fs_fonts *sf, int property)
{
  LOGFONT *lf;

  lf = FS_GET_LOGFONT(sf, property);
  if (!lf) {
    switch (property) {
    case 0:
      lf = &default_normal_font;
      break;
    case 1:
      lf = &default_bold_font;
      break;
    case 2:
      lf = &default_italic_font;
      break;
    case 3:
      lf = &default_italic_bold_font;
      break;
    default :
      return (LPLOGFONT) NULL;
    }
    lf->lfWidth = fs -> hmetric;
    lf->lfHeight = fs -> vmetric;
  }
  return lf;
}

HFONT
get_hfont(win32_fontset *fs, fs_fonts *sf, int property)
{
  LOGFONT *lf;
  HFONT hf;

  lf = get_logfont(fs, sf, property);
  hf = CreateFontIndirect(lf);
  FS_GET_FONT(sf, property) = hf;
  return hf;
}

/* patch 96.11.17  by K.Fujii */
HFONT
win32_fs_load_font (win32_fontset *fs, int lc, int property,
		    Lisp_Object *coding_system)
{
  fs_fonts **sf;
  int fsc = (lc & 0x7f);

  sf = fs->fonts;
  if (sf [fsc] != NULL) {
    if (! FS_GET_FONT (sf[fsc], property)) {
      HFONT hf;
      LOGFONT *lf = get_logfont (fs, sf[fsc], property);
      hf = CreateFontIndirect (lf);
      FS_GET_FONT (sf[fsc], property) = hf;
    }
    *coding_system = sf[fsc]->coding_system;
    return FS_GET_FONT (sf[fsc], property);
  }
  return (HFONT) NULL;
}
/* end of patch */

XFontStruct *
get_font_type(unsigned char *name,
	     fs_fonts* sf, int property)
{
  XFontStruct *result;
  int nbytes;

  result = (XFontStruct *)xmalloc(sizeof(XFontStruct));
  nbytes = strlen(name) + 1;
  result -> name = (unsigned char *) xmalloc(nbytes);
  memcpy(result -> name, name, nbytes);
  result -> phfont = &FS_GET_FONT(sf, property);
  result -> property_index = property;
  result -> fs_font = sf;

  return result;
}

XFontStruct *
win32_load_font(struct frame *f, unsigned char *name, int property)
{
  fs_fonts **sf;
  LOGFONT *lf;
  HFONT hf;
  int i;

  sf = FRAME_FONTSET(f) -> fonts;
  for (i = 0; i < 128; i++){
    if (i == (LCJP     & 0x7f) || i == (LCROMAN & 0x7f) ||
	i == (LCJPOLD  & 0x7f) || i == (LCJP2   & 0x7f) ||
	i == (LCBIG5_2 & 0x7f))
      continue;
    if (sf [i]) {
      if (strcmp(sf[i]->name, name) == 0) {
	hf = FS_GET_FONT(sf [i], property);
	if (hf) return get_font_type(name, sf [i], property);
	hf = get_hfont(FRAME_FONTSET(f), sf [i], property);
	if (!hf) return (XFontStruct *) NULL;
	return get_font_type(name, sf [i], property);
      }
    }
  }
  return (XFontStruct *) NULL;
}

void
win32_unload_font(struct frame *f, XFontStruct *font)
{
  HFONT *phf;

  phf =  &FS_GET_FONT(font->fs_font, font -> property_index);

  if (!(*phf)) return;
/*  DeleteObject(*phf);
  *phf = NULL; */
  xfree(font->name);
  xfree(font);

  return;
}

win32_fontset *
win32_get_fontset(unsigned char *name)
{
  int i;

  for (i = 0;i < fontset_num; i++) {
    if (strcmp(win32_fontset_table[i] -> fs_name, name) == 0) {
      return win32_fontset_table[i];
    }
  }
  return 0;
}

void
expand_fontset_table(int num)
{
  int i, j;
  win32_fontset *fs;

  if (! win32_fontset_table) {
    win32_fontset_table = (win32_fontset **) xmalloc (num * sizeof (win32_fontset *));
    fontset_num = 0;
  } else
    win32_fontset_table = (win32_fontset **) 
      xrealloc(win32_fontset_table, 
	       (fontset_num + num) * sizeof(win32_fontset *));
  fs = (win32_fontset *) xmalloc(num * sizeof(win32_fontset));
  for (i = 0; i < num; i++) {
    win32_fontset_table[fontset_num] = fs++;
    for (j = 0; j < 128; j++)
      win32_fontset_table[fontset_num]->fonts [j] = NULL;
    fontset_num++;
  }
}

void
internal_add_fontset(int location,
		     const unsigned char *name,
		     int width, int height, int base, int overhang)
{
  win32_fontset *tp;
  tp = win32_fontset_table[location];
  tp -> fs_name = (unsigned char *) xmalloc(strlen(name) + 1);
  strcpy(tp -> fs_name, name);
  tp -> vmetric = height;
  tp -> hmetric = width;
  tp -> font_base = base;
  tp -> overhang = overhang;
}

void
internal_add_font(win32_fontset *fontset,
		  const unsigned char *name,
		  Lisp_Object coding_system,
		  int lc)
{
  fs_fonts *fp;
  int i;
  int fsc;

  if (lc == LCJP2) lc = LCKANA;
  fsc = (lc & 0x7f);

  Fcheck_code (coding_system);

  if (fontset->fonts [fsc] == NULL) {
    fp = fontset->fonts [fsc] = (fs_fonts *) xmalloc (sizeof (fs_fonts));
    bzero (fp, sizeof (fs_fonts));
    fp->name = (unsigned char *) xmalloc (strlen (name) + 1);
  } else {
    fp = (fontset -> fonts)[fsc];
    fp->name = (unsigned char *) xrealloc (fp->name, strlen (name) + 1);
    for (i = 0; i < FONT_PROPERTY_NUM; i++) {
      if (fp->font_properties [i]) xfree (fp->font_properties[i]);
      fp -> font_properties[i] = (HFONT) 0;
      if (fp->logfont_properties [i]) xfree (fp->logfont_properties [i]);
      fp -> logfont_properties[i] = (LPLOGFONT) 0;
    }
  }
  strcpy(fp -> name, name);
  /* fp->fsCharSet = fsc; */
  fp->coding_system = coding_system;
  if (lc == LCKANA  || lc == LCROMAN ||
      lc == LCJPOLD || lc == LCJP) {
    static int jp_fsc[] = { LCKANA & 0x7f, LCROMAN & 0x7f, LCJPOLD & 0x7f, LCJP & 0x7f};
    int i;

    for (i = 0; i < 4; i++) {
      if (fsc != jp_fsc [i])
	fontset->fonts [jp_fsc [i]] = fp;
    }
  } else if (lc == LCBIG5_1) {
    fontset->fonts [LCBIG5_2 & 0x7f] = fp;
  } else if (lc == LCBIG5_2) {
    fontset->fonts [LCBIG5_1 & 0x7f] = fp;
  }
}

/* patch 96.11.2  by K.Fujii */
Lisp_Object Qansi_charset, Qdefault_charset, Qsymbol_charset, Qshiftjis_charset, Qoem_charset;
Lisp_Object Qhangeul_charset, Qgb2312_charset, QchaineseBig5_charset;
Lisp_Object Qiso8859_1_charset, Qiso8859_charset;
Lisp_Object Qascii_charset, Qlatin1_charset, Qjapanese_charset, Qkorean_charset, Qchinese_charset;
Lisp_Object Qiso2022_jp_charset, Qiso2022_kr_charset, Qiso2022_cn_charset;
Lisp_Object Qhebrew_charset, Qarabic_charset, Qgreek_charset, Qturkish_charset;
Lisp_Object Qthai_charset, Qcyrillic_charset;
Lisp_Object Qheavy_weight, Qextrabold_weight, Qbold_weight, Qsemibold_weight, Qmedium_weight;
Lisp_Object Qnormal_weight, Qlight_weight, Qextralight_weight, Qthin_weight;
/* end of patch */

Lisp_Object
logfont_to_lisp_object(LOGFONT *lf)
{
  Lisp_Object logfont[9];
  /* [0] is name.
     [1] is width. 
     [2] is height. 
     [3] is weight.
     [4] is orientation.
     [5] is italic.
     [6] is strikeout.
     [7] is charset.
     [8] is quality. 
     [9] is OutPrecision 
     [10] is PitchAndFamily */

  Lisp_Object found_code;
  unsigned char *buf;
  int nbytes, size;
  coding_type mccode;

  nbytes = strlen (lf->lfFaceName);
  encode_code (Fcheck_code (Vwin32_system_coding_system), &mccode);
  CODE_CNTL (&mccode) |= CC_END;
  buf = (unsigned char *)xmalloc (ENCODE_BUF_SIZE (nbytes, &mccode));
  size = encode (&mccode, lf->lfFaceName, buf, nbytes, &found_code);
  logfont[0] = make_string(buf, size);
  xfree(buf);

  logfont[1] = make_number(lf->lfWidth);
  logfont[2] = make_number(lf->lfHeight);
#if 1  /* patch 96.11.2  by K.Fujii */
  if      (lf->lfWeight >= FW_HEAVY)      logfont[3] = Qheavy_weight;
  else if (lf->lfWeight >= FW_EXTRABOLD)  logfont[3] = Qextrabold_weight;
  else if (lf->lfWeight >= FW_BOLD)       logfont[3] = Qbold_weight;
  else if (lf->lfWeight >= FW_SEMIBOLD)   logfont[3] = Qsemibold_weight;
  else if (lf->lfWeight >= FW_MEDIUM)     logfont[3] = Qmedium_weight;
  else if (lf->lfWeight >= FW_NORMAL)     logfont[3] = Qnormal_weight;
  else if (lf->lfWeight >= FW_LIGHT)      logfont[3] = Qlight_weight;
  else if (lf->lfWeight >= FW_EXTRALIGHT) logfont[3] = Qextralight_weight;
  else if (lf->lfWeight >= FW_THIN)       logfont[3] = Qthin_weight;
  else                                    XSETFASTINT (logfont[3], lf->lfWeight);
  XSETFASTINT(logfont[4], 0);
#else
  XSETFASTINT(logfont[3], lf->lfWeight);
  XSETFASTINT(logfont[4], lf->lfOrientation);
#endif
  logfont[5] = (lf->lfItalic) ? Qt:Qnil;
  logfont[6] = (lf->lfStrikeOut) ? Qt:Qnil;
#if 1  /* patch 96.11.2  by K.Fujii */
  switch (lf->lfCharSet)
    {
    case ANSI_CHARSET:        logfont[7] = Qansi_charset;         break;
    case DEFAULT_CHARSET:     logfont[7] = Qdefault_charset;      break;
    case SYMBOL_CHARSET:      logfont[7] = Qsymbol_charset;       break;
    case SHIFTJIS_CHARSET:    logfont[7] = Qjapanese_charset;     break;
    case HANGEUL_CHARSET:     logfont[7] = Qkorean_charset;       break;
    case GB2312_CHARSET:      logfont[7] = Qchinese_charset;      break;
    case CHINESEBIG5_CHARSET: logfont[7] = QchaineseBig5_charset; break;
    case HEBREW_CHARSET:      logfont[7] = Qhebrew_charset;       break;
    case ARABIC_CHARSET:      logfont[7] = Qarabic_charset;       break;
    case GREEK_CHARSET:       logfont[7] = Qgreek_charset;        break;
    case TURKISH_CHARSET:     logfont[7] = Qturkish_charset;      break;
    case THAI_CHARSET:        logfont[7] = Qthai_charset;         break;
    case RUSSIAN_CHARSET:     logfont[7] = Qcyrillic_charset;     break;
    case OEM_CHARSET:         logfont[7] = Qoem_charset;          break;
    default:                  XSETFASTINT (logfont[7], lf->lfCharSet);
    }
#else  
  XSETFASTINT(logfont[7], lf->lfCharSet);
#endif
  XSETFASTINT(logfont[8], lf->lfQuality);
  XSETFASTINT(logfont[9], lf->lfOutPrecision);
  XSETFASTINT(logfont[10], lf->lfPitchAndFamily);

  return Flist(XFASTINT(11), logfont);
}

LOGFONT lisp_object_to_logfont(Lisp_Object font)
{
  coding_type mccode;
  LOGFONT lf;
  Lisp_Object tmpcdr, tmpcar;
  unsigned char *buf;
  int size;

  tmpcar = Fcar(font);
  encode_code(Fcheck_code(Vwin32_system_coding_system), &mccode);

  buf = alloca(DECODE_BUF_SIZE(XSTRING(tmpcar) -> size +1, &mccode));

  /* LOGFONT is NEVER accessed by lisp interpreter. */
  CODE_CNTL (&mccode) |= CC_END;
  size = decode(&mccode, XSTRING(tmpcar) -> data, buf,
		XSTRING(tmpcar) -> size);
  if (size >= LOGFONT_FACENAME_MAX) {
    error ("Font name is too long. ");
   }
  memcpy(lf.lfFaceName, buf, size);
  lf.lfFaceName[size] = (TCHAR) NULL;

  tmpcdr = Fcdr(font);
  tmpcar = Fcar(tmpcdr);
  lf.lfWidth = XINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfHeight = XINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
#if 1  /* patch 96.11.2  by K.Fujii */
  if      (EQ (tmpcar, Qheavy_weight))      lf.lfWeight = FW_HEAVY;
  else if (EQ (tmpcar, Qextrabold_weight))  lf.lfWeight = FW_EXTRABOLD;
  else if (EQ (tmpcar, Qbold_weight))       lf.lfWeight = FW_BOLD;
  else if (EQ (tmpcar, Qsemibold_weight))   lf.lfWeight = FW_SEMIBOLD;
  else if (EQ (tmpcar, Qmedium_weight))     lf.lfWeight = FW_MEDIUM;
  else if (EQ (tmpcar, Qnormal_weight))     lf.lfWeight = FW_NORMAL;
  else if (EQ (tmpcar, Qlight_weight))      lf.lfWeight = FW_LIGHT;
  else if (EQ (tmpcar, Qextralight_weight)) lf.lfWeight = FW_EXTRALIGHT;
  else if (EQ (tmpcar, Qthin_weight))       lf.lfWeight = FW_THIN;
  else
#endif
  lf.lfWeight = XFASTINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfOrientation = XFASTINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfItalic = !NILP(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfStrikeOut = !NILP(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
#if 1  /* patch 96.11.2  by K.Fujii */
  if      (EQ (tmpcar, Qansi_charset) ||
           EQ (tmpcar, Qiso8859_1_charset) ||
           EQ (tmpcar, Qiso8859_charset))      lf.lfCharSet = ANSI_CHARSET;
  else if (EQ (tmpcar, Qhebrew_charset))       lf.lfCharSet = HEBREW_CHARSET;
  else if (EQ (tmpcar, Qarabic_charset))       lf.lfCharSet = ARABIC_CHARSET;
  else if (EQ (tmpcar, Qgreek_charset))        lf.lfCharSet = GREEK_CHARSET;
  else if (EQ (tmpcar, Qturkish_charset))      lf.lfCharSet = TURKISH_CHARSET;
  else if (EQ (tmpcar, Qthai_charset))         lf.lfCharSet = THAI_CHARSET;
  else if (EQ (tmpcar, Qcyrillic_charset))     lf.lfCharSet = RUSSIAN_CHARSET;
  else if (EQ (tmpcar, Qdefault_charset))      lf.lfCharSet = DEFAULT_CHARSET;
  else if (EQ (tmpcar, Qsymbol_charset))       lf.lfCharSet = SYMBOL_CHARSET;
  else if (EQ (tmpcar, Qshiftjis_charset) ||
	   EQ (tmpcar, Qiso2022_jp_charset) ||
	   EQ (tmpcar, Qjapanese_charset))     lf.lfCharSet = SHIFTJIS_CHARSET;
  else if (EQ (tmpcar, Qhangeul_charset) ||
	   EQ (tmpcar, Qiso2022_kr_charset) ||
	   EQ (tmpcar, Qkorean_charset))       lf.lfCharSet = HANGEUL_CHARSET;
  else if (EQ (tmpcar, Qgb2312_charset) ||
	   EQ (tmpcar, Qiso2022_cn_charset) ||
	   EQ (tmpcar, Qchinese_charset))      lf.lfCharSet = GB2312_CHARSET;
  else if (EQ (tmpcar, QchaineseBig5_charset)) lf.lfCharSet = CHINESEBIG5_CHARSET;
  else if (EQ (tmpcar, Qoem_charset))          lf.lfCharSet = OEM_CHARSET;
  else
#endif
  lf.lfCharSet = XFASTINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfQuality = XFASTINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfOutPrecision = XFASTINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfPitchAndFamily = XFASTINT(tmpcar);

/* Set the default value... */
  lf.lfEscapement = lf.lfOrientation;
  lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
  lf.lfUnderline = 0;

  return lf;
}

DEFUN("win32-check-font", Fwin32_check_font, Swin32_check_font, 1, 1, 0,
      "Check validity of WIN32-FONT. \n\
The validity of WIN32-FONT is it is a list that have 11 elements.\n\
(name width height weight orientation italic strikeout charset quality\n\
OutPrecision PitchAndFamily)")
  (font)
     Lisp_Object font;
/* The width must be a string that describes the name of its font.\n\
The height must be a number that describes the width of its font.\n\
The weight must be a number that describes the weight.(1-1000)\n\
The orientation must be a number that describes the orientation.(angle)\n\
If the italic is nil, it is not italic font.\n\
If the strikeout is nil, it doesn't have a strikeout line.\n\
The charset must be a number that describes the charset that it uses.\n\
(128 is SJIS, 0 is ANSI)\n\
The quality must be a number that describes its quality.(0-2) */
{
  Lisp_Object tmpcar, tmpcdr;
  CHECK_LIST(font, 0);
  tmpcar = Fcar(font);
  tmpcdr = Fcdr(font);
  CHECK_STRING(tmpcar, 1); /* name */
  if ((XSTRING(tmpcar)->size) > LOGFONT_FACENAME_MAX) {
    error("The name is too long. Max is %d", LOGFONT_FACENAME_MAX);
  }
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar,2); /* width */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar,3); /* height */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
#if 1  /* patch 96.11.2  by K.Fujii */
  if (EQ (tmpcar, Qheavy_weight) ||
      EQ (tmpcar, Qextrabold_weight) ||
      EQ (tmpcar, Qbold_weight) ||
      EQ (tmpcar, Qsemibold_weight) ||
      EQ (tmpcar, Qmedium_weight) ||
      EQ (tmpcar, Qnormal_weight) ||
      EQ (tmpcar, Qlight_weight) ||
      EQ (tmpcar, Qextralight_weight) ||
      EQ (tmpcar, Qthin_weight))
    ;
  else
#endif
  CHECK_NUMBER(tmpcar, 4); /* weight */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar, 5); /* orientation */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  if (!((EQ(tmpcar, Qnil)) || (EQ(tmpcar, Qt)))) { /* italic */
    error("italic cell must be nil or t");
  }
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  if (!((EQ(tmpcar, Qnil)) || (EQ(tmpcar, Qt)))) { /* strikeout */
    error("strikeout cell must be nil or t");
  }
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
#if 1  /* patch 96.11.2  by K.Fujii */
  if (EQ (tmpcar, Qansi_charset) ||
      EQ (tmpcar, Qiso8859_1_charset) ||
      EQ (tmpcar, Qiso8859_charset) ||
      EQ (tmpcar, Qhebrew_charset) ||
      EQ (tmpcar, Qarabic_charset) ||
      EQ (tmpcar, Qgreek_charset) ||
      EQ (tmpcar, Qturkish_charset) ||
      EQ (tmpcar, Qthai_charset) ||
      EQ (tmpcar, Qcyrillic_charset) ||
      EQ (tmpcar, Qdefault_charset) ||
      EQ (tmpcar, Qsymbol_charset) ||
      EQ (tmpcar, Qshiftjis_charset) ||
      EQ (tmpcar, Qjapanese_charset) ||
      EQ (tmpcar, Qiso2022_jp_charset) ||
      EQ (tmpcar, Qhangeul_charset) ||
      EQ (tmpcar, Qkorean_charset) ||
      EQ (tmpcar, Qiso2022_kr_charset) ||
      EQ (tmpcar, Qgb2312_charset) ||
      EQ (tmpcar, Qchinese_charset) ||
      EQ (tmpcar, QchaineseBig5_charset) ||
      EQ (tmpcar, Qiso2022_cn_charset) ||
      EQ (tmpcar, Qoem_charset))
    ;
  else
#endif
  CHECK_NUMBER(tmpcar, 6); /* charset */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar, 7); /* quality */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar, 8); /* OutPrecision */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar, 9); /* PitchAndFamily */
  if (!NILP(tmpcdr)) {
    error ("invalid font. its length must be 11.");
  }
  return Qt;
}

void
win32_load_font_tmp(win32_fontset *fs, XFontStruct *font, int forcedef)
{
  fs_fonts **sf;
  LOGFONT *lf;
  HFONT hf;
  int i;

  sf = fs -> fonts;
  for (i = 0; i < 128; i++){
    if (i == (LCJP     & 0x7f) || i == (LCROMAN & 0x7f) ||
	i == (LCJPOLD  & 0x7f) || i == (LCJP2   & 0x7f) ||
	i == (LCBIG5_2 & 0x7f))
      continue;
    if (sf [i]) {
      if (strcmp(sf [i]->name, font->name) == 0) {
	font->fs_font = sf [i];
	hf = FS_GET_FONT(sf [i], font->property_index);
	if (hf) {
	  font->phfont = &FS_GET_FONT(sf [i], font->property_index);
	  return;
	}
	hf = get_hfont(fs, sf [i], font->property_index);
	if (!hf) {
	  error("Unable to change fontset. Fonts in the fontset you specified are not found.");
	}
	font->phfont = &FS_GET_FONT(sf [i], font->property_index);
	return;
      }
    }
  }

 error:
#if 0
  error("Fail to change fontset!. There is no font that is in the previous fontset in the specified fontset.");
#else
  if (!forcedef)
    win32_load_font_tmp(fs, font, 1);
  else
  error("Fail to change fontset!. There is no font that is in the previous fontset in the specified fontset. Furturemore, there is no default font!.");
#endif
}

/* patch 96.12.7  by K.Fujii */
void
adjusting_fontset_font_size (struct frame *f, win32_fontset *fs)
{
  int max_width, max_height, max_overhang;
  int i;
  TEXTMETRIC tm;
  HDC hdc;
  HANDLE oldobj;
  HFONT hf;

  max_width    = fs->hmetric;
  max_height   = fs->vmetric;
  max_overhang = fs->overhang;

  hdc = my_get_dc (FRAME_WIN32_WINDOW (f));
  for (i = 0; i < 128; i++) {
    if (i == (LCJP     & 0x7f) || i == (LCROMAN & 0x7f) ||
	i == (LCJPOLD  & 0x7f) || i == (LCJP2   & 0x7f) ||
	i == (LCBIG5_2 & 0x7f))
      continue;
    if ( fs->fonts [i] != NULL) {
      BOOL v;

      if (hf = get_hfont (fs, fs->fonts [i], 0)) {
	oldobj = SelectObject (hdc, hf);
	v = GetTextMetrics (hdc, &tm);
	SelectObject (hdc, oldobj);
	if (v != FALSE) {
	  if (max_width < tm.tmAveCharWidth)
	    max_width = tm.tmAveCharWidth;
	  if (max_height < tm.tmHeight)
	    max_height = tm.tmHeight;
	  if (max_overhang < tm.tmOverhang)
	    max_overhang = tm.tmOverhang;
	}
      }
    }
  }

  fs->hmetric = max_width;
  fs->vmetric = max_height;
  fs->overhang = max_overhang;
  ReleaseDC (FRAME_WIN32_WINDOW (f), hdc);
}
/* end of patch */

void
win32_change_fontset(struct frame *f, unsigned char *name)
{

  win32_fontset *fs;
  int i;
  HWND hwnd;

  fs = win32_get_fontset(name);
  if (!fs) {
    error ("Such fontset does not exist!");
  }

  /* patch 96.12.7  by K.Fujii */
  adjusting_fontset_font_size (f, fs);
  /* end of patch */

  /* display.win32 member is initialized by bzero */
  if (f->output_data.win32->font)
    win32_load_font_tmp(fs, f->output_data.win32->font, 0);
  for (i = 0; i < FRAME_N_PARAM_FACES(f); i++)
    if (FRAME_PARAM_FACES(f)[i] != 0
	&& FRAME_PARAM_FACES(f)[i]->font != (XFontStruct *) FACE_DEFAULT)
      {
/*	win32_unload_font_tmp(FRAME_FONTSET(f),
			      f->display.win32->param_faces[i]->font); */
	win32_load_font_tmp(fs, FRAME_PARAM_FACES(f)[i]->font, 0);
      }
  FRAME_FONTSET(f) = fs;
  frame_update_line_height (f);
  x_set_window_size (f, 0, f->width, f->height);
  hwnd = FRAME_WIN32_WINDOW(f);
  if (hwnd){
#if 1 /* patch 96.11.16  by K.Fujii */
    SendMessage (hwnd, WM_IME_SETFONT,
		 (WPARAM) get_logfont (fs,
				       f->output_data.win32->font->fs_font, 
				       f->output_data.win32->font->property_index),
		 0);
#else
    win32_set_ime_font(hwnd, 
		       get_logfont(fs,
				   f->output_data.win32->font->fs_font, 
				   f->output_data.win32->font->
				   property_index));
#endif
  }
}

/* fontset manager */
DEFUN("win32-add-fontset", Fwin32_add_fontset, Swin32_add_fontset, 5, 5, 0,
      "Add fontset. First ARG is a name of fontset.\n\
Second ARG, Third ARG, Forth ARG, and Fifth ARG are \n\
metric infomation, width, height, base, and overhang.")
(name, width, height, base, overhang)
     Lisp_Object name, width, height, base, overhang;
{
  win32_fontset *fs;

  CHECK_STRING(name, 0);
  CHECK_NUMBER(width, 1);
  CHECK_NUMBER(height, 1);
  CHECK_NUMBER(base, 1);
  CHECK_NUMBER(overhang, 1);
  fs = win32_get_fontset (XSTRING (name)->data);
  if (! fs) {
    expand_fontset_table(1);
    internal_add_fontset(fontset_num - 1, XSTRING(name) -> data, 
			 XINT (width), XINT (height), XINT (base), XINT (overhang));
  } else if (fs->vmetric != XINT (width) ||
	     fs->hmetric != XINT (height) ||
	     fs->font_base != XINT (base) ||
	     fs->overhang != XINT (overhang)) {
    int i, j;
    fs_fonts *fp;

    fs->vmetric = XINT (width);
    fs->hmetric = XINT (height);
    fs->font_base = XINT (base);
    fs->overhang = XINT (overhang);
    for (i = 0; i < 128; i++) {
      if (i == (LCJP     & 0x7f) || i == (LCROMAN & 0x7f) ||
	  i == (LCJPOLD  & 0x7f) || i == (LCJP2   & 0x7f) ||
	  i == (LCBIG5_2 & 0x7f) ||
	  fs->fonts [i] == NULL)
	continue;
      fp = fs->fonts [i];
      for (j = 0; j < FONT_PROPERTY_NUM; j++) {
	if (fp->font_properties [j])
	  xfree (fp->font_properties);
	if (fp->logfont_properties [j])
	  xfree (fp->logfont_properties [j]);
      }
      xfree (fp->name);
      xfree (fp);
      fs->fonts [i] = NULL;
    }
  }

  return Qnil;
}

DEFUN("win32-fontset-list", Fwin32_fontset_list, Swin32_fontset_list, 0, 0, 0,
      "Return a list of all win32 fontset")
  ()
{
  Lisp_Object val = Qnil;
  int i;

  for (i = 0;i < fontset_num;i++) {
    val = Fcons(build_string(win32_fontset_table[i] -> fs_name), val);
  }
  return val;
}

DEFUN("win32-refresh-fontset", Fwin32_refresh_fontset,
      Swin32_refresh_fontset, 1, 1, 0,
      "Refresh the fontset of the frame.\n\
FRAME shows the frame you want to refresh fontset.")
  (frame)
     Lisp_Object frame;
{
  CHECK_FRAME(frame, 0);

  win32_change_fontset(XFRAME(frame),
		       FRAME_FONTSET(XFRAME(frame)) -> fs_name);
  return Qnil;
}

DEFUN("win32-change-fontset-attribute", Fwin32_change_fontset_attribute,
      Swin32_change_fontset_attribute, 4, 5, 0,
      "Change the fontset attribute.\n\
The first arg WIDTH is the width of the fontset. The second arg HEIGHT is\n\
the height of it.  The third arg BASE is the base length. The forth arg \n\
OVERHANG is the overhang. The Fifth optional arg NAME is the name of the \n\
fontset you want to change. If it is nil, the fontset of the current frame\n\
is changed.")
  (width, height, base, overhang, name)
     Lisp_Object width, height, base, overhang, name;
{
  win32_fontset *fs;

  CHECK_NUMBER(width, 0);
  CHECK_NUMBER(height, 1);
  CHECK_NUMBER(overhang, 2);
  CHECK_NUMBER(base, 3);

  if (NILP(name)) {
    fs = FRAME_FONTSET(selected_frame);
  }else{
    CHECK_STRING(name, 4);
    fs = win32_get_fontset(XSTRING(name)->data);
    if (!fs) {
      error("Such fontset does NOT exist!.");
    }
  }
  fs->hmetric = XFASTINT(width);
  fs->vmetric = XFASTINT(height);
  fs->overhang = XFASTINT(overhang);
  fs->font_base = XFASTINT(base);

  return Qnil;
}

/* font manager */
DEFUN ("win32-add-font", Fwin32_add_font, Swin32_add_font, 3, 4, 0,
       "Add font. The first ARG is FONTNAME(string) of font.\n\
The second ARG is the LEADING-CHAR(integer) that is use with its font.\n\
The third ARG is CODING-SYSTEM(coding-system) that is this use with its font.\n\
The fourth optional ARG is a FONTSETNAME(string) than this font is added to.")
  (name, lc, coding_system, fontset)
     Lisp_Object name, lc, coding_system, fontset;
{
  win32_fontset *fs;
  int font_location;

  /* patch 96.11.9, 96.11.18  by K.Fujii */
  if (SYMBOLP (lc) && ! NILP (Fboundp (lc)))
    lc = Fsymbol_value (lc);
  CHECK_NUMBER (lc, 1);
  if (XFASTINT (lc) >= 256 || char_type[XFASTINT (lc)] == TYPEINV)
    error ("Invalid leading-char: %d", XFASTINT (lc));
  /* end of patch */

  CHECK_STRING(name, 0);

  if (NILP(fontset)){
    fs = CURRENT_FONTSET;
  }else{
    CHECK_STRING(fontset, 1);
    fs = win32_get_fontset(XSTRING(fontset) -> data);
    if (!fs) {
      error ("Such fontset does not exist!");
    }
  }

  internal_add_font(fs,
		    XSTRING(name) -> data,
		    coding_system,
		    XINT (lc));
  return Qnil;
}

DEFUN("win32-font-list", Fwin32_font_list, Swin32_font_list, 0, 1, 0,
      "Return a list of all win32 font. \n\
The optional arg is the fontset where you want to get fonts.")
(fontset)
{
  win32_fontset *fs;
  int i;
  Lisp_Object val = Qnil;

  if (NILP(fontset)){
    fs = CURRENT_FONTSET;
  }else{
    CHECK_STRING(fontset, 0);
    fs = win32_get_fontset(XSTRING(fontset) -> data);
    if (!fs) {
      error ("Such fontset does not exist!");
    }
  }
  for(i = 0; i < 128; i++) {
    if (i == (LCJP     & 0x7f) || i == (LCROMAN & 0x7f) ||
	i == (LCJPOLD  & 0x7f) || i == (LCJP2   & 0x7f) ||
	i == (LCBIG5_2 & 0x7f))
      continue;
    if ((fs->fonts) [i])
      val = Fcons(build_string((fs->fonts)[i] -> name), val);
  }
  return val;
}
  
/* font property manager */
DEFUN("win32-change-font-property", Fwin32_change_font_property,
      Swin32_change_font_property, 3, 4, 0,
      "Change font property. \n\
The first arg is a name of the font where the property is changed.\n\
The second ARG is a number of the property.\n\
The third ARG is the logfont infomation.\n\
The forth optional arg is a name of the fontset \n\
where this property is changed.")
(font, number, logfont, fontset)
{
  win32_fontset *fs;
  fs_fonts **sf;
  LPLOGFONT plf;
  int i, prop_num;
  CHECK_STRING(font, 0);
  CHECK_NUMBER(number, 1);
  prop_num = XFASTINT(number);
  if (prop_num >= FONT_PROPERTY_NUM) {
    error("Too large font property number.");
  }

  if (NILP(fontset)){
    fs = CURRENT_FONTSET;
  }else{
    CHECK_STRING(fontset, 2);
    fs = win32_get_fontset(XSTRING(fontset) -> data);
    if (!fs) {
      error ("Such fontset does not exist!");
    }
  }
  Fwin32_check_font(logfont);

  sf = fs -> fonts;
  for (i = 0; i < 128; i++){
    if (i == (LCJP     & 0x7f) || i == (LCROMAN & 0x7f) ||
	i == (LCJPOLD  & 0x7f) || i == (LCJP2   & 0x7f) ||
	i == (LCBIG5_2 & 0x7f))
      continue;
    if (sf [i]) {
      if (strcmp(sf [i]->name, XSTRING(font)->data) == 0) {
	plf = FS_GET_LOGFONT(sf [i], XFASTINT(number));
	if (!plf) {
	  plf = (LPLOGFONT)xmalloc(sizeof(LOGFONT));
	  FS_GET_LOGFONT(sf [i], XFASTINT(number)) = plf;
	}
	*plf = lisp_object_to_logfont(logfont);
	return Qnil;
      }
    }
  }
  error ("Such font does not exist!");
}

DEFUN ("win32-get-font-property", Fwin32_get_font_property, 
       Swin32_get_font_property, 2, 3, 0,
       "Get the property of the font you specified.\n\
The first arg is a name of the font.\n\
The second arg is a number of the property.\n\
The third optional arg is a fontset of the font.")
(font, number, fontset)
{
  win32_fontset *fs;
  fs_fonts **sf;
  int i, prop_num;
  LOGFONT *lf;

  CHECK_STRING(font, 0);
  CHECK_NUMBER(number, 1);
  prop_num = XFASTINT(number);
  if (prop_num >= FONT_PROPERTY_NUM) {
    error("Too large font property number.");
  }

  if (NILP(fontset)){
    fs = CURRENT_FONTSET;
  }else{
    CHECK_STRING(fontset, 2);
    fs = win32_get_fontset(XSTRING(fontset) -> data);
    if (!fs) {
      error ("Such fontset does not exist!");
    }
  }

  sf = fs -> fonts;
  for (i = 0; i < 128; i++){
    if (i == (LCJP     & 0x7f) || i == (LCROMAN & 0x7f) ||
	i == (LCJPOLD  & 0x7f) || i == (LCJP2   & 0x7f) ||
	i == (LCBIG5_2 & 0x7f))
      continue;
    if (sf [i]) {
      if (strcmp(sf [i]-> name, XSTRING(font)->data) == 0) {
	if (lf = FS_GET_LOGFONT(sf [i], XFASTINT(number))) {
	  return logfont_to_lisp_object(lf);
	}
	return Qnil;
      }
    }
  }
  error ("Such font does not exist!");
}

DEFUN ("win32-get-font-metric", Fwin32_get_font_metric, 
       Swin32_get_font_metric, 2, 3, 0,
       "Get the metric of the font you specified.\n\
The first arg is a name of the font.\n\
The second arg is a number of the property.\n\
The third optional arg is a fontset of the font.\n\
The returned metric infomation is as follows.\n\
(width, height, base, overhang)")
  (font, number, fontset)
     Lisp_Object font, number, fontset;
{
  win32_fontset *fs;
  fs_fonts **sf;
  int i, prop_num;
  HFONT hf;

  CHECK_STRING(font, 0);
  CHECK_NUMBER(number, 1);
  prop_num = XFASTINT(number);
  if (prop_num >= FONT_PROPERTY_NUM) {
    error("Too large font property number.");
  }

  if (NILP(fontset)){
    fs = CURRENT_FONTSET;
  }else{
    CHECK_STRING(fontset, 2);
    fs = win32_get_fontset(XSTRING(fontset) -> data);
    if (!fs) {
      error ("Such fontset does not exist!");
    }
  }

  sf = fs -> fonts;
  for (i = 0; i < 128; i++){
    if (i == (LCJP     & 0x7f) || i == (LCROMAN & 0x7f) ||
	i == (LCJPOLD  & 0x7f) || i == (LCJP2   & 0x7f) ||
	i == (LCBIG5_2 & 0x7f))
      continue;
    if (sf [i]) {
      if (strcmp(sf [i]-> name, XSTRING(font)->data) == 0) {
	if (hf = get_hfont(fs, sf [i], XFASTINT(number))) {
	  HDC hdc;
	  HANDLE oldobj;
	  TEXTMETRIC tm;

	  hdc = my_get_dc (FRAME_WIN32_WINDOW(selected_frame));
	  oldobj = SelectObject(hdc, hf);
	  if (!GetTextMetrics(hdc, &tm)) {
	    error("Fail to get metrics!!");
	  }
	  SelectObject(hdc, oldobj);
	  ReleaseDC(FRAME_WIN32_WINDOW(selected_frame), hdc);
	  return Fcons(make_number((int)tm.tmAveCharWidth), 
		       Fcons(make_number((int)tm.tmHeight),
			     Fcons(make_number((int)tm.tmAscent),
				   Fcons(make_number((int)tm.tmOverhang), 
					 Qnil))));
	}
	return Qnil;
      }
    }
  }
  error ("Such font does not exist!");
}

DEFUN ("win32-query-get-logfont", Fwin32_query_get_logfont,
       Swin32_query_get_logfont, 0, 0, 0,
       "Return the font that you selected with Font Select Dialog Box.")
     ()
{
  CHOOSEFONT cf;
  LOGFONT lf;
  unsigned char buf[100];

  cf.lStructSize = sizeof(CHOOSEFONT);
  cf.hwndOwner = FRAME_WIN32_WINDOW(selected_frame);
  cf.hDC = (HDC)NULL;
  cf.lpLogFont = &lf;
  cf.Flags = CF_SCREENFONTS | CF_FIXEDPITCHONLY | 
    CF_FORCEFONTEXIST | CF_EFFECTS;
  cf.rgbColors = RGB(0,0,0);
  cf.lpfnHook = (LPCFHOOKPROC)NULL;
  cf.lpTemplateName = (LPSTR)NULL;
  cf.hInstance = (HINSTANCE) NULL;
  cf.lpszStyle = (LPSTR)NULL;
  cf.nFontType = SCREEN_FONTTYPE;

  ChooseFont(&cf);
  return logfont_to_lisp_object(&lf);
}

/* syms */

syms_of_win32fontset()
{
  DEFVAR_LISP("win32-system-coding-system",
	      &Vwin32_system_coding_system,
	      "coding system in win32 system");
  /* patch 96.11.9  by K.Fujii */
  DEFVAR_LISP ("WIN32-FONTSET", &Vwin32_fontset, "");
  Vwin32_fontset = Qt;
  /* end of patch */
  defsubr(&Swin32_check_font);
  defsubr(&Swin32_change_font_property);
  defsubr(&Swin32_add_font);
  defsubr(&Swin32_add_fontset);
  defsubr(&Swin32_refresh_fontset);
  defsubr(&Swin32_change_fontset_attribute);
  defsubr(&Swin32_check_font);
  defsubr(&Swin32_fontset_list);
  defsubr(&Swin32_font_list);
  defsubr(&Swin32_get_font_property);
  defsubr(&Swin32_get_font_metric);
  defsubr(&Swin32_query_get_logfont);
/*
   defsubr(&Swin32_delete_fontset);
   defsubr(&Swin32_delete_font);
   defsubr(&Swin32_get_fontset_info);
   defsubr(&Swin32_get_font_info);
*/
  /* patch 96.11.2, 96.11.7  by K.Fujii */
  Qascii_charset = intern ("ascii");
  staticpro (&Qascii_charset);
  Qansi_charset = intern ("ansi");
  staticpro (&Qansi_charset);
  Qiso8859_1_charset = intern ("iso8859-1");
  staticpro (&Qiso8859_1_charset);
  Qiso8859_charset = intern ("iso8859");
  staticpro (&Qiso8859_charset);
  Qlatin1_charset = intern ("latin-1");
  staticpro (&Qlatin1_charset);
  Qhebrew_charset = intern ("Hebrew");
  staticpro (&Qhebrew_charset);
  Qarabic_charset = intern ("Arabic");
  staticpro (&Qarabic_charset);
  Qgreek_charset = intern ("Greek");
  staticpro (&Qgreek_charset);
  Qturkish_charset = intern ("Turkish");
  staticpro (&Qturkish_charset);
  Qthai_charset = intern ("Thai");
  staticpro (&Qthai_charset);
  Qcyrillic_charset = intern ("Cyrillic");
  staticpro (&Qcyrillic_charset);
  Qdefault_charset = intern ("default");
  staticpro (&Qdefault_charset);
  Qsymbol_charset = intern ("symbol");
  staticpro (&Qsymbol_charset);
  Qshiftjis_charset = intern ("shift-JIS");
  staticpro (&Qshiftjis_charset);
  Qjapanese_charset = intern ("Japanese");
  staticpro (&Qjapanese_charset);
  Qiso2022_jp_charset = intern ("iso-2022-jp");
  staticpro (&Qiso2022_jp_charset);
  Qhangeul_charset = intern ("hangeul");
  staticpro (&Qhangeul_charset);
  Qkorean_charset = intern ("Korean");
  staticpro (&Qkorean_charset);
  Qiso2022_kr_charset = intern ("iso-2022-kr");
  staticpro (&Qiso2022_kr_charset);
  Qgb2312_charset = intern ("gb2312");
  staticpro (&Qgb2312_charset);
  Qiso2022_cn_charset = intern ("iso-2022-cn");
  staticpro (&Qiso2022_cn_charset);
  QchaineseBig5_charset = intern ("big5");
  staticpro (&QchaineseBig5_charset);
  Qchinese_charset = intern ("Chinese");
  staticpro (&Qchinese_charset);
  Qoem_charset = intern ("oem");
  staticpro (&Qoem_charset);

  Qheavy_weight = intern ("heavy");
  staticpro (&Qheavy_weight);
  Qextrabold_weight = intern ("extrabold");
  staticpro (&Qextrabold_weight);
  Qbold_weight = intern ("bold");
  staticpro (&Qbold_weight);
  Qsemibold_weight = intern ("semibold");
  staticpro (&Qsemibold_weight);
  Qmedium_weight = intern ("medium");
  staticpro (&Qmedium_weight);
  Qnormal_weight = intern ("normal");
  staticpro (&Qnormal_weight);
  Qlight_weight = intern ("light");
  staticpro (&Qlight_weight);
  Qextralight_weight = intern ("extralight");
  staticpro (&Qextralight_weight);
  Qthin_weight = intern ("thin");
  staticpro (&Qthin_weight);
  /* end of patch */
}
