/* Win32 specific Lisp staff.
   Copyright (C) 1992, 1995 Free Software Foundation, Inc.

This file is part of Mule (MULtilingual Enhancement of GNU Emacs).

Mule is free software distributed in the form of patches to GNU Emacs.
You can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

Mule is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* 93.6.1 written by M.Higashida <manabu@sigmath.osaka-u.ac.jp> */
/* 95.7.23 modified by himi for mule 2.2.2 console window */

/* Big chenged for Mule-2.3@19.30(Win32) 19996  by K.Fujii */

#include "config.h"

#ifdef NULL
#undef NULL
#endif
#include "lisp.h"

#ifdef HAVE_NTGUI
#include "frame.h"
#include "w32term.h"
#include "buffer.h"
#include "window.h"
#else
#include <windows.h>
#endif

#ifdef IME_CONTROL
#include <winnls32.h>
#include <ime.h>
#include <dde.h>
#endif /* IME_CONTROL */
#include "ntheap.h"

#ifdef IME_CONTROL

#ifndef CFS_FORCE_POSITION
#define IMC_SETCOMPOSITIONWINDOW        0x000C
typedef struct tagCOMPOSITIONFORM {
    DWORD dwStyle;
    POINT ptCurrentPos;
    RECT  rcArea;
} COMPOSITIONFORM;
#define CFS_FORCE_POSITION           0x20
#define CFS_POINT                    0x02
#define CFS_RECT		     0x01
#endif

BOOL fIME = FALSE;
typedef LRESULT (WINAPI *SENDIMEMESSAGEEXPROC)(HWND, LPARAM);
SENDIMEMESSAGEEXPROC SendIMEMessageExProc;
typedef HWND (WINAPI *IMMGETDEFAULTIMEWNDPROC)(HWND);
IMMGETDEFAULTIMEWNDPROC ImmGetDefaultIMEWndProc;
HANDLE hIME;
int inhibit_fep_control = 1;
int ime_inited;
#endif /* IME_CONTROL */


#ifdef IME_CONTROL

int
win32_ime_control_init (void)
{
  HMODULE hImm32;
  HMODULE hUser32;

  fIME = FALSE;
  SendIMEMessageExProc = NULL;
  ImmGetDefaultIMEWndProc = NULL;

  hUser32 = GetModuleHandle ("USER32.DLL");
  if (is_this_platform_version4 ())
    hImm32  = GetModuleHandle ("IMM32.DLL");

  if (hUser32)
    {
      SendIMEMessageExProc =
	(SENDIMEMESSAGEEXPROC) GetProcAddress (hUser32, "SendIMEMessageExA");
      if (hImm32 != NULL)
	ImmGetDefaultIMEWndProc =
	  (IMMGETDEFAULTIMEWNDPROC) GetProcAddress (hImm32,
						    "ImmGetDefaultIMEWnd");

      if (SendIMEMessageExProc) {
	fIME = TRUE;
	if ((hIME = GlobalAlloc (GMEM_SHARE | GMEM_MOVEABLE, sizeof(IMESTRUCT))) == NULL) {
	  fIME = FALSE;
	  SendIMEMessageExProc = NULL;
	  ImmGetDefaultIMEWndProc = NULL;
	  return 0;
	}
	return 1;
      }
    }
  return 0;
}

int
win32_ime_getopen (HWND hwnd)
{
  if (fIME) {
    int result = 0;
    LPIMESTRUCT lpIme = (LPIMESTRUCT) GlobalLock (hIME);
    lpIme->fnc = IME_GETOPEN;

    if (! (SendIMEMessageExProc) (hwnd, (LPARAM) hIME))
      result = -1;
    else if (lpIme->wCount > 0)
      result = lpIme->wCount;

    GlobalUnlock (hIME);
    return result;
  }
  return 0;
}

void
win32_ime_setopen (HWND hwnd, int open)
{
  if (fIME) {
    LPIMESTRUCT lpIme = (LPIMESTRUCT) GlobalLock (hIME);
    lpIme->fnc = IME_SETOPEN;
    lpIme->wParam = open;
    (SendIMEMessageExProc) (hwnd, (LPARAM) hIME);
    GlobalUnlock (hIME);
  }
}
    
#ifdef WIN32_FONTSET
void
win32_set_ime_font(HWND hwnd, LPLOGFONT psetlf)
{
  if (fIME)
    {
      LPIMESTRUCT lpIme = (LPIMESTRUCT) GlobalLock (hIME);
      HANDLE hplf;
      LPLOGFONT plf;
      
      lpIme->fnc = IME_SETCONVERSIONFONTEX;
      hplf = GlobalAlloc (GMEM_SHARE | GMEM_MOVEABLE, sizeof(LOGFONT));
      plf = (LPLOGFONT) GlobalLock(hplf);

      *plf = *psetlf;
      lpIme -> lParam1 = (LPARAM)hplf;

      (SendIMEMessageExProc)(hwnd, (LPARAM) hIME);

      GlobalUnlock (hplf);
      GlobalUnlock (hIME);
      GlobalFree (hplf);
    }
}
#endif  

void
win32_set_ime_conv_window(HWND hwnd, int x, int y)
{
  if (fIME) {
    LONG size, units;
    int top, left, bottom, right;

    enter_crit ();
    units = GetWindowLong (hwnd, WND_UNITS_INDEX);
    size = GetWindowLong (hwnd, WND_FRAME_SIZE_INDEX);
    top = left = (int) GetWindowWord (hwnd, WND_INTERNAL_BORDER_WIDTH_INDEX);
    leave_crit ();
      
    right  = (left + ((LOWORD (size) - 1) * LOWORD (units))) + 1;
    bottom = (top  + ((HIWORD (size) - 1) * HIWORD (units))) + 1;
    if (x < left) x = left;
    if (x > right) x = right;
    if (y < top) y = top;
    if (y > bottom) y = bottom;

    if (!ImmGetDefaultIMEWndProc) {
      LPIMESTRUCT lpIme = (LPIMESTRUCT) GlobalLock (hIME);

      lpIme->fnc     = IME_SETCONVERSIONWINDOW;
      lpIme->wParam  = MCW_WINDOW | MCW_RECT;
      lpIme->lParam1 = MAKELRESULT (x, y);
      lpIme->lParam2 = MAKELRESULT (left, top);
      lpIme->lParam3 = MAKELRESULT (right, bottom);
      GlobalUnlock (hIME);
      (SendIMEMessageExProc) (hwnd, (LPARAM) hIME);
    } else {
      HWND IMEhwnd;
      COMPOSITIONFORM compform;

      IMEhwnd = (ImmGetDefaultIMEWndProc)(hwnd);

      compform.dwStyle = CFS_RECT;
      compform.ptCurrentPos.x = x;
      compform.ptCurrentPos.y = y;
      compform.rcArea.top = top;
      compform.rcArea.left = left;
      compform.rcArea.bottom = bottom;
      compform.rcArea.right = right;

      SendMessage(IMEhwnd, WM_IME_CONTROL, (WPARAM)IMC_SETCOMPOSITIONWINDOW,
		  (LPARAM)(&compform));
    }
    enter_crit ();
    SetWindowLong (hwnd, WND_IME_CONVERT_POS_INDEX, MAKELRESULT (x, y));
    leave_crit ();
  }
}

DEFUN ("fep-get-mode", Ffep_get_mode, Sfep_get_mode, 0, 1, 0, "")
     (frame)
     Lisp_Object frame;
{
  if (! inhibit_fep_control && ime_inited) {
    struct frame *f = FRAMEP (frame) ? XFRAME (frame) : selected_frame;

    return (SendMessage ((HWND)(FRAME_WIN32_WINDOW (f)),
			 WM_IME_GETOPEN, 0, 0) >= 0
	    ? Qt : Qnil);
  }
  return Qnil;
}

DEFUN ("fep-force-on", Ffep_force_on, Sfep_force_on, 0, 1, 0, "")
     (frame)
     Lisp_Object frame;
{
  if (! inhibit_fep_control && ime_inited) {
    struct frame *f = FRAMEP (frame) ? XFRAME (frame) : selected_frame;

    SendMessage ((HWND)(FRAME_WIN32_WINDOW (f)),
		 WM_IME_SETOPEN, (WPARAM) IWP_ON, 0);
  }
  return Qnil;
}

DEFUN ("fep-force-off", Ffep_force_off, Sfep_force_off, 0, 1, 0, "")
     (frame)
     Lisp_Object frame;
{
  if (! inhibit_fep_control && ime_inited) {
    struct frame *f = FRAMEP (frame) ? XFRAME (frame) : selected_frame;

    SendMessage ((HWND)(FRAME_WIN32_WINDOW (f)),
		 WM_IME_SETOPEN, (WPARAM) IWP_OFF, 0);
  }
  return Qnil;
}

DEFUN ("does-ime-have-undetermined-strings",
       Fdoes_ime_have_undetermined_strings,
       Sdoes_ime_have_undetermined_strings, 0, 1, 0, "")
     (frame)
     Lisp_Object frame;
{
  if (! inhibit_fep_control && ime_inited) {
    struct frame *f = (FRAMEP (frame) ? XFRAME (frame) : selected_frame);
    HWND hwnd = FRAME_WIN32_WINDOW (f);

    return  (SendMessage (hwnd, WM_IME_GETOPEN, 0, 0) > 0 ? Qt : Qnil);
  }
  return Qnil;
}

DEFUN ("fep-init", Ffep_init, Sfep_init, 0, 0, "", "")
     ()
{
  if (! inhibit_fep_control && ! ime_inited) {
    Lisp_Object ime_id_string;
    MSG msg;
    extern DWORD dwWinThreadId;

#if 0
    PostThreadMessage (dwWinThreadId, WM_IME_INIT, 0, 0);
    GetMessage (&msg, NULL, WM_EMACS_DONE, WM_EMACS_DONE);
#endif

    if (fIME) {
      if (ImmGetDefaultIMEWndProc != NULL)
	ime_id_string = build_string ("IMM32");
      else if (SendIMEMessageExProc != NULL)
	ime_id_string = build_string ("USER32");
      else
	ime_id_string = Qnil;

      if (! NILP (ime_id_string)) {
	ime_inited = 1;
      }
      return ime_id_string;
    }
  }
  return Qnil;
}

#endif /* IME_CONTROL */


syms_of_win32ime()
{
#ifdef IME_CONTROL
  DEFVAR_BOOL ("inhibit-fep-control", &inhibit_fep_control, "");
  inhibit_fep_control = 1;
  defsubr (&Sfep_force_on);
  defsubr (&Sfep_force_off);
  defsubr (&Sfep_init);
  defsubr (&Sfep_get_mode);
  defsubr (&Sdoes_ime_have_undetermined_strings);
#endif /* IME_CONTROL */
}
